(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2010-2018)
Author:  Julian Nagele <julian.nagele@uibk.ac.at> (2012-2017)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2010-2016)
Author:  Sarah Winkler <sarah.winkler@uibk.ac.at> (2013, 2014)
Author:  Thomas Sternagel <thomas.sternagel@uibk.ac.at> (2012)
Author:  Akihisa Yamada <akihisa.yamada@uibk.ac.at> (2017)
License: LGPL (see file COPYING.LESSER)
*)
theory CPF_Parser
  imports
    Proof_Checker
    XTC_Parser
    LTS.LTS_Parser
begin

hide_const (open) Congruence.partition
hide_const (open) Invariant_Checker.name
hide_const (open) AC_Subterm_Criterion.proj_term
hide_const (open) Finite_Cartesian_Product.mat
hide_const (open) Simplex.Bounds
hide_const (open) Cooperation_Program.sharp.Sharp
hide_type (open) Finite_Cartesian_Product.vec
hide_const (open) Congruence.eq
hide_const (open) UnivPoly.deg

abbreviation xml_singleton :: "ltag \<Rightarrow> 'a xmlt2 \<Rightarrow> ('a \<Rightarrow> 'b) \<Rightarrow> 'b xmlt2"
where "xml_singleton tag p f \<equiv> XMLdo tag {x \<leftarrow> p; xml_return (f x)}"

abbreviation xml_pair
where "xml_pair tag p1 p2 f \<equiv> XMLdo tag {a \<leftarrow> p1; b \<leftarrow> p2; xml_return (f a b)}"

abbreviation xml_triple :: "ltag \<Rightarrow> 'a xmlt2 \<Rightarrow> 'b xmlt2 \<Rightarrow> 'c xmlt2 \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> 'c \<Rightarrow> 'd) \<Rightarrow> 'd xmlt2"
where "xml_triple tag p1 p2 p3 f \<equiv> XMLdo tag {a \<leftarrow> p1; b \<leftarrow> p2; c \<leftarrow> p3; xml_return (f a b c)}"

abbreviation xml_many where "xml_many tag p f \<equiv> XMLdo tag {a \<leftarrow>* p; xml_return (f a)}"
abbreviation xml_tuple4 where "xml_tuple4 tag p1 p2 p3 p4 f \<equiv>
  XMLdo tag {a \<leftarrow> p1; b \<leftarrow> p2; c \<leftarrow> p3; d \<leftarrow> p4; xml_return (f a b c d)}"
abbreviation xml_tuple5 where "xml_tuple5 tag p1 p2 p3 p4 p5 f \<equiv>
  XMLdo tag {a \<leftarrow> p1; b \<leftarrow> p2; c \<leftarrow> p3; d \<leftarrow> p4; e \<leftarrow> p5; xml_return (f a b c d e)}"
abbreviation "xml_tuple6 tag p1 p2 p3 p4 p5 p6 t \<equiv>
 XMLdo tag {a \<leftarrow> p1; b \<leftarrow> p2; c \<leftarrow> p3; d \<leftarrow> p4; e \<leftarrow> p5; f \<leftarrow> p6; xml_return (t a b c d e f)}"

partial_function (sum_bot) arith_fun :: "arithFun xmlt2"
  where
  [code]: "arith_fun xml = XMLdo (STR ''arithFunction'') {
    a \<leftarrow>
      xml_change (xml_nat (STR ''natural'')) (\<lambda> n. xml_return (Const n))
      XMLor xml_change (xml_nat (STR ''variable'')) (\<lambda> n. xml_return (Arg (n - 1)))
      XMLor XMLdo (STR ''sum'') { as \<leftarrow>* arith_fun; xml_return (Sum as)}
      XMLor XMLdo (STR ''product'') { as \<leftarrow>* arith_fun; xml_return (Prod as)}
      XMLor XMLdo (STR ''max'') { as \<leftarrow>* arith_fun; xml_return (Semantic_Labeling_Carrier.Max as)}
      XMLor xml_many (STR ''min'') arith_fun Semantic_Labeling_Carrier.Min \<comment> \<open>just illustrate other style\<close>
      XMLor XMLdo (STR ''ifEqual'') { a \<leftarrow> arith_fun; b \<leftarrow> arith_fun; c \<leftarrow> arith_fun; d  \<leftarrow> arith_fun;
         xml_return (IfEqual a b c d)}
    ;
    xml_return a
  } xml"

definition sl_variant :: "'f xmlt2 \<Rightarrow> ('f, string) sl_variant xmlt2"
where
  "sl_variant xml2name = xml_singleton (STR ''model'') (
     XMLdo (STR ''rootLabeling'') { fo \<leftarrow>? xml2name;
        xml_return (Rootlab (case fo of None \<Rightarrow> None | Some f \<Rightarrow> Some (f,1))) }
     XMLor XMLdo (STR ''finiteModel'') {
        n \<leftarrow> xml_change (xml_nat (STR ''carrierSize'')) (\<lambda> n. xml_return (n - 1));
        to \<leftarrow>? xml_singleton (STR ''tupleOrder'') (xml_leaf (STR ''pointWise'') ()) id;
        inter \<leftarrow>* xml_triple (STR ''interpret'')
           xml2name
           (xml_nat (STR ''arity''))
           arith_fun
           (\<lambda> f a inter. ((f,a),inter));
        xml_return (case to of
          None \<Rightarrow> Finitelab (SL_Inter n inter)
        | Some _ \<Rightarrow> QuasiFinitelab (SL_Inter n inter) [])
     }
   ) id"
hide_const (open) sl_variant

context
  fixes C :: "'a ring" (structure)
begin
fun sum_lpoly :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a list" where
  "sum_lpoly [] ys = ys"
| "sum_lpoly xs [] = xs"
| "sum_lpoly (x # xs) (y # ys) = (x \<oplus> y) # sum_lpoly xs ys"

fun
  lpoly_of ::
    "(nat, 'a) tpoly \<Rightarrow> String.literal +\<^sub>\<bottom> ('a \<times> 'a list)"
where
  "lpoly_of (PNum i) = return (i, [])"
| "lpoly_of (PVar x) = return (\<zero>, replicate x \<zero> @ [\<one>])"
| "lpoly_of (PSum []) = return (\<zero>, [])"
| "lpoly_of (PSum (p # ps)) = (
     lpoly_of  p
  \<bind> (\<lambda>(cp, ncp). lpoly_of (PSum ps)
  \<bind> (\<lambda>(cq, ncq). return (cp \<oplus> cq, sum_lpoly ncp ncq))))"
| "lpoly_of (PMult []) = return (\<one>, [])"
| "lpoly_of (PMult (p # ps)) = (
     lpoly_of p
  \<bind> (\<lambda>(cp, ncp). lpoly_of (PMult ps)
  \<bind> (\<lambda>(cq, ncq). if (Ball (set ncp) ((=) \<zero>))
       then return (cp \<otimes> cq, map ((\<otimes>) cp) ncq)
       else (if (Ball (set ncq) ((=) \<zero>))
         then return (cp \<otimes> cq, map (\<lambda> x. x \<otimes> cq) ncp)
         else error (STR ''cannot transform non-linear polynomial to linear polynomial'')))))"

primrec fit_length :: "nat \<Rightarrow> 'a list \<Rightarrow> 'a list" where
  "fit_length (Suc n) as = (case as of
    [] \<Rightarrow> replicate (Suc n) \<zero>
  | b # bs \<Rightarrow> b # fit_length n bs)"
| "fit_length 0 _ = []"
end

partial_function (sum_bot) polynomial :: "'a xmlt2 \<Rightarrow> (nat, 'a) tpoly xmlt2" where
  [code]: "polynomial xml2coeff x = xml_singleton (STR ''polynomial'') (
       xml_singleton (STR ''coefficient'') xml2coeff PNum
       XMLor xml_change (xml_nat (STR ''variable'')) (\<lambda> n. xml_return (PVar (n - 1)))
       XMLor xml_many (STR ''sum'') (polynomial xml2coeff) PSum
       XMLor xml_many (STR ''product'') (polynomial xml2coeff) PMult
     ) id x"

definition "int_coeff = xml_int (STR ''integer'')"
hide_const (open) int_coeff

definition vec_coeff :: "'a xmlt2 \<Rightarrow> 'a vec xmlt2" where
  "vec_coeff xml2coeff = xml_many (STR ''vector'')
     (xml_singleton (STR ''coefficient'') xml2coeff id) vec_of_list"
hide_const (open) vec_coeff

(* allow vectors and convert them to matrices by filling in zeros *)
definition mat_coeff :: "nat \<Rightarrow> 'a \<Rightarrow> 'a xmlt2 \<Rightarrow> 'a mat xmlt2" where
  "mat_coeff n ze xml2coeff =
     xml_many (STR ''matrix'') (CPF_Parser.vec_coeff xml2coeff) (mat_of_cols n)
     XMLor xml_change (CPF_Parser.vec_coeff xml2coeff) (\<lambda> v. xml_return (mat n n (\<lambda> (i,j). if j = 0 then v $ i else ze)))"
hide_const (open) mat_coeff

definition arctic_coeff :: "arctic xmlt2" where
  "arctic_coeff = xml_change CPF_Parser.int_coeff (xml_return o Num_arc)
     XMLor xml_leaf (STR ''minusInfinity'') MinInfty"
hide_const (open) arctic_coeff

definition xml_rat :: "rat xmlt2"
where
  "xml_rat = xml_change (xml_int (STR ''integer'')) (xml_return o of_int)
     XMLor xml_pair (STR ''rational'')
        (xml_int (STR ''numerator''))
        (xml_int (STR ''denominator''))
        (\<lambda> x y. of_int x / of_int y)"

definition arctic_rat_coeff :: "rat arctic_delta xmlt2" where
  "arctic_rat_coeff = xml_change xml_rat (xml_return o Num_arc_delta)
       XMLor xml_leaf (STR ''minusInfinity'') MinInfty_delta"

hide_const (open) arctic_rat_coeff

datatype Domain =
  Natural nat | Integer | Arctic | Arctic_rat | Int_mat nat nat | Arctic_mat nat
| Arctic_rat_mat nat | Rational rat nat | Rat_mat nat nat | Mini_Alg real nat
| Mini_Alg_mat nat nat

definition rat_domain :: "rat xmlt2" where
  "rat_domain = xml_singleton (STR ''rationals'') (xml_singleton (STR ''delta'') xml_rat id) id"
hide_const (open) rat_domain

definition xml_real :: "real xmlt2" where
  "xml_real = xml_change xml_rat (xml_return o real_of_rat)
      XMLor xml_triple (STR ''algebraic'') xml_rat xml_rat xml_rat
        (\<lambda>a b c. (real_of_rat a + real_of_rat b * sqrt (real_of_rat c)))"

definition real_domain :: "real xmlt2" where
  "real_domain = xml_singleton (STR ''algebraicNumbers'') (xml_singleton (STR ''delta'') xml_real id) id"
hide_const (open) real_domain

definition basic_domain :: "(nat \<Rightarrow> Domain) xmlt2" where
  "basic_domain = xml_leaf (STR ''naturals'') Natural
    XMLor xml_leaf (STR ''integers'') (\<lambda> d. Integer)
    XMLor xml_change CPF_Parser.rat_domain (xml_return o Rational)
    XMLor xml_change CPF_Parser.real_domain (xml_return o Mini_Alg)
    XMLor xml_singleton (STR ''arctic'') (xml_singleton (STR ''domain'') (
      xml_leaf (STR ''naturals'') (\<lambda> d. Arctic)
      XMLor xml_leaf (STR ''integers'') (\<lambda> d. Arctic)
      XMLor xml_change CPF_Parser.rat_domain (\<lambda> r. xml_return (\<lambda> d. Arctic_rat))
     ) id) id"
hide_const (open) basic_domain

definition interpretation_type :: "Domain xmlt2" where
  "interpretation_type = xml_singleton (STR ''type'') (
     xml_triple (STR ''matrixInterpretation'')
       (xml_singleton (STR ''domain'') CPF_Parser.basic_domain id)
       (xml_nat (STR ''dimension''))
       (xml_nat (STR ''strictDimension''))
       (\<lambda>domain di sd. case domain 0 of
         Natural _ \<Rightarrow> Int_mat di sd
       | Integer \<Rightarrow> Int_mat di sd
       | Rational _ _  \<Rightarrow> Rat_mat di sd
       | Mini_Alg _ _  \<Rightarrow> Mini_Alg_mat di sd
       | Arctic \<Rightarrow> Arctic_mat di
       | Arctic_rat \<Rightarrow> Arctic_rat_mat di
       )
    XMLor xml_pair (STR ''polynomial'')
       (xml_singleton (STR ''domain'') (
          CPF_Parser.basic_domain
          XMLor xml_triple (STR ''matrices'')
            (xml_nat (STR ''dimension''))
            (xml_nat (STR ''strictDimension''))
            (xml_singleton (STR ''domain'') CPF_Parser.basic_domain id)
            (\<lambda>di sd domain d. case domain d of
               Natural _ \<Rightarrow> Int_mat di sd
             | Rational _ _ \<Rightarrow> Rat_mat di sd
             | Mini_Alg _ _ \<Rightarrow> Mini_Alg_mat di sd
             | Integer \<Rightarrow> Int_mat di sd
             | Arctic \<Rightarrow> Arctic_mat di
             | Arctic_rat \<Rightarrow> Arctic_rat_mat di)
          ) id)
       (xml_nat (STR ''degree''))
       (\<lambda>type d. type d)
   ) id"
hide_const (open) interpretation_type

type_synonym ('f,'a) inter = "('f \<times> nat) \<times> 'a \<times> 'a list"
type_synonym ('f,'a) mat_inter = "('f \<times> nat) \<times> 'a mat \<times> 'a mat list"
type_synonym ('f,'a) nl_inter = "('f \<times> nat) \<times> (nat, 'a) poly"

datatype (dead 'f) "interpretation" =
  (* linear polynomials (with support of negative constants *)
  Int_linear_poly (int_linear_poly: "('f,int)inter")
| Rat_linear_poly  (rat_linear_poly: "('f,rat)inter")
| Arctic_linear_poly (arctic_linear_poly:  "('f,arctic)inter")
| Arctic_rat_linear_poly  (arctic_rat_linear_poly: "('f,rat arctic_delta)inter")
| Real_linear_poly  (real_linear_poly: "('f,real)inter")
  (* linear polynomials over matrices (with support of negative constants *)
| Int_matrix  (int_matrix: "('f,int)mat_inter")
| Rat_matrix  (rat_matrix: "('f,rat)mat_inter")
| Arctic_matrix  (arctic_matrix: "('f,arctic)mat_inter")
| Arctic_rat_matrix  (arctic_rat_matrix: "('f,rat arctic_delta)mat_inter")
| Real_matrix  (real_matrix: "('f,real)mat_inter")
  (* non linear polynomials (without support of negative constants) *)
| Int_non_linear_poly  (int_non_linear_poly: "('f,int)nl_inter")
| Rat_non_linear_poly  (rat_non_linear_poly: "('f,rat)nl_inter")
| Real_non_linear_poly  (real_non_linear_poly: "('f,real)nl_inter")

(* the flag bi is to indicate that we are interested in bounded increase,
   where rationals always have to be chosen as non-linear variant *)
abbreviation class_ring where
  "class_ring t \<equiv> class_semiring t ()"

abbreviation mat_ring :: "('a :: semiring_1) itself \<Rightarrow> nat \<Rightarrow> 'a mat ring" where
  "mat_ring t n \<equiv> ring_mat t n ()"

definition "xml_change' f p x = bind2 (f p) (\<lambda> e. xml_error e x) Right"

definition "interpretation" :: "bool \<Rightarrow> 'a xmlt2 \<Rightarrow> 'a redtriple_impl xmlt2" where
  "interpretation bi xml2name = XMLdo (STR ''interpretation'') {
       type \<leftarrow> CPF_Parser.interpretation_type;
       pi \<leftarrow>* XMLdo (STR ''interpret'') {f \<leftarrow> xml2name; a \<leftarrow> (xml_nat (STR ''arity''));
         case type of
           Integer \<Rightarrow> XMLdo {
             poly \<leftarrow> xml_change (CPF_Parser.polynomial CPF_Parser.int_coeff) (xml_return o poly_of);
             xml_return (Int_non_linear_poly ((f,a),poly))}
         | Natural deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then let c = class_ring (TYPE(int)) in
              (XMLdo {
               poly \<leftarrow> xml_change (CPF_Parser.polynomial CPF_Parser.int_coeff) (xml_change' (lpoly_of c));
               xml_return (Int_linear_poly ((f,a),(fst poly, fit_length c a (snd poly))))})
              else
              XMLdo {
               poly \<leftarrow> xml_change (CPF_Parser.polynomial CPF_Parser.int_coeff) (xml_return o poly_of);
               xml_return (Int_non_linear_poly ((f,a),poly))}
         | Rational d deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then let c = class_ring (TYPE(rat)) in
              (XMLdo {
               poly \<leftarrow> xml_change (CPF_Parser.polynomial xml_rat) (xml_change' (lpoly_of c));
               xml_return (Rat_linear_poly ((f,a),(fst poly, fit_length c a (snd poly))))})
              else
              XMLdo {
               poly \<leftarrow> xml_change (CPF_Parser.polynomial xml_rat) (xml_return o poly_of);
               xml_return (Rat_non_linear_poly ((f,a),poly))}
         | Mini_Alg d deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then let c = class_ring (TYPE(real)) in
              (XMLdo {
               poly \<leftarrow> xml_change (CPF_Parser.polynomial xml_real) (xml_change' (lpoly_of c));
               xml_return (Real_linear_poly ((f,a),(fst poly, fit_length c a (snd poly))))})
              else
              XMLdo {
               poly \<leftarrow> xml_change (CPF_Parser.polynomial xml_real) (xml_return o poly_of);
               xml_return (Real_non_linear_poly ((f,a),poly))}
         | Arctic \<Rightarrow> let c = class_ring (TYPE(arctic)) in
              (XMLdo {
                 poly \<leftarrow> xml_change (CPF_Parser.polynomial CPF_Parser.arctic_coeff) (xml_change' (lpoly_of c));
                 xml_return (Arctic_linear_poly ((f,a),(fst poly, fit_length c a (snd poly))))})
         | Arctic_rat \<Rightarrow> let c = class_ring (TYPE(rat arctic_delta)) in
              (XMLdo {
                 poly \<leftarrow> xml_change (CPF_Parser.polynomial CPF_Parser.arctic_rat_coeff) (xml_change' (lpoly_of c));
                 xml_return (Arctic_rat_linear_poly ((f,a),(fst poly, fit_length c a (snd poly))))})
         | Int_mat n sd \<Rightarrow> let c = mat_ring (TYPE(int)) n in
              (XMLdo {
                 poly \<leftarrow> xml_change (CPF_Parser.polynomial(CPF_Parser.mat_coeff n 0 CPF_Parser.int_coeff)) (xml_change' (lpoly_of c));
                 xml_return (Int_matrix ((f,a),(fst poly, fit_length c a (snd poly))))})
         | Rat_mat n sd \<Rightarrow> let c = mat_ring (TYPE(rat)) n in
              (XMLdo {
                 poly \<leftarrow> xml_change (CPF_Parser.polynomial(CPF_Parser.mat_coeff n 0 xml_rat)) (xml_change' (lpoly_of c));
                 xml_return (Rat_matrix ((f,a),(fst poly, fit_length c a (snd poly))))})
         | Mini_Alg_mat n sd \<Rightarrow> let c = mat_ring (TYPE(real)) n in
              (XMLdo {
                 poly \<leftarrow> xml_change (CPF_Parser.polynomial(CPF_Parser.mat_coeff n 0 xml_real)) (xml_change' (lpoly_of c));
                 xml_return (Real_matrix ((f,a),(fst poly, fit_length c a (snd poly))))})
         | Arctic_mat n \<Rightarrow> let c = mat_ring (TYPE(arctic)) n in
              (XMLdo {
                 poly \<leftarrow> xml_change (CPF_Parser.polynomial(CPF_Parser.mat_coeff n 0 CPF_Parser.arctic_coeff)) (xml_change' (lpoly_of c));
                 xml_return (Arctic_matrix ((f,a),(fst poly, fit_length c a (snd poly))))})
         | Arctic_rat_mat n \<Rightarrow> let c = mat_ring (TYPE(rat arctic_delta)) n in
              (XMLdo {
                 poly \<leftarrow> xml_change (CPF_Parser.polynomial(CPF_Parser.mat_coeff n 0 CPF_Parser.arctic_rat_coeff)) (xml_change' (lpoly_of c));
                 xml_return (Arctic_rat_matrix ((f,a),(fst poly, fit_length c a (snd poly))))})
       };
       xml_return (case type of
            Natural deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then Int_carrier (map int_linear_poly pi) else Int_nl_carrier (map int_non_linear_poly pi)
          | Integer \<Rightarrow> Int_nl_carrier (map int_non_linear_poly pi)
          | Rational d deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then Rat_carrier (map rat_linear_poly pi) else Rat_nl_carrier d (map rat_non_linear_poly pi)
          | Mini_Alg d deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then Real_carrier (map real_linear_poly pi) else Real_nl_carrier d (map real_non_linear_poly pi)
          | Arctic \<Rightarrow> Arctic_carrier (map arctic_linear_poly pi)
          | Arctic_rat \<Rightarrow> Arctic_rat_carrier (map arctic_rat_linear_poly pi)
          | Int_mat n sd \<Rightarrow> Int_mat_carrier n sd (map int_matrix pi)
          | Rat_mat n sd \<Rightarrow> Rat_mat_carrier n sd (map rat_matrix pi)
          | Mini_Alg_mat n sd \<Rightarrow> Real_mat_carrier n sd (map real_matrix pi)
          | Arctic_mat n \<Rightarrow> Arctic_mat_carrier n (map arctic_matrix pi)
          | Arctic_rat_mat n \<Rightarrow> Arctic_rat_mat_carrier n (map arctic_rat_matrix pi)
      )
    }
  "

hide_const (open) "interpretation"

definition bounds_type :: "boundstype xmlt2" where
  "bounds_type = xml_singleton (STR ''type'') (
     xml_leaf (STR ''roof'') Roof
     XMLor xml_leaf (STR ''match'') Match
   ) id"

definition bounds_bound :: "nat xmlt2" where
  "bounds_bound = xml_nat (STR ''bound'')"

definition state :: "string xmlt2" where
  "state = xml_text (STR ''state'')"

definition final_states :: "string list xmlt2" where
  "final_states = XMLdo (STR ''finalStates'') {ret \<leftarrow>* state; xml_return ret}"

definition ta_bounds_lhs where
  "ta_bounds_lhs xml2name = XMLdo (STR ''lhs'') {
    XMLdo {
      a \<leftarrow> state;
      xml_return (Inr a)
    } XMLor XMLdo {
      f \<leftarrow> xml2name;
      h \<leftarrow> xml_nat (STR ''height'');
      qs \<leftarrow>* state;
      xml_return (Inl ((f, h), qs))
    }
  }"

definition ta_normal_lhs where
  "ta_normal_lhs xml2name = XMLdo (STR ''lhs'') {
    XMLdo {
      a \<leftarrow> state;
      xml_return (Inr a)
    } XMLor XMLdo {
      a \<leftarrow> xml2name;
      b \<leftarrow>* state;
      xml_return (Inl (a,b))
    }
  }"

definition transition where
  "transition xml2lhs = xml_pair (STR ''transition'')
    xml2lhs
    (xml_singleton (STR ''rhs'') state id)
    (\<lambda>lhs q. case lhs of
      Inl (f, qs) \<Rightarrow> Inl (TA_rule f qs q)
    | Inr q' \<Rightarrow> Inr (q', q))"

definition "transitions xml2lhs = XMLdo (STR ''transitions'') {
  rls \<leftarrow>* transition xml2lhs;
  let (rules, eps) = partition (\<lambda>rl. case rl of Inl _ \<Rightarrow> True | Inr _ \<Rightarrow> False) rls;
  let ruls = map (\<lambda>rl. case rl of Inl r \<Rightarrow> r) rules;
  let ep   = map (\<lambda>rl. case rl of Inr e \<Rightarrow> e) eps;
  xml_return (ruls, ep)
}"

definition "tree_automaton xml2lhs = xml_pair (STR ''treeAutomaton'')
  final_states
  (transitions xml2lhs)
  (\<lambda>f (t, eps). Tree_Automaton f t eps)"

definition closed_criterion :: "string ta_relation xmlt2" where
 "closed_criterion = XMLdo (STR ''criterion'') {
    ret \<leftarrow>
      xml_leaf (STR ''compatibility'') Id_Relation
      XMLor xml_singleton (STR ''stateCompatibility'') (
        XMLdo (STR ''relation'') {
          ret \<leftarrow>* xml_pair (STR ''entry'') state state Pair;
          xml_return ret
        }
      ) Some_Relation
      XMLor xml_leaf (STR ''decisionProcedure'') Decision_Proc
      XMLor xml_leaf (STR ''decisionProcedureOld'') Decision_Proc_Old;
    xml_return ret
  }"

definition bounds_info :: "'a xmlt2 \<Rightarrow> ('a, string) bounds_info xmlt2"
  where
    "bounds_info xml2name = XMLdo (STR ''bounds'') {
      a \<leftarrow> bounds_type;
      b \<leftarrow> bounds_bound;
      c \<leftarrow> final_states;
      d \<leftarrow> tree_automaton (ta_bounds_lhs xml2name);
      e \<leftarrow>[Id_Relation] closed_criterion;
      xml_return (Bounds_Info a b c d e)
    }"

context
  fixes xml2name :: "'a :: showl xmlt2"
begin

partial_function (sum_bot) ctxt ::  "('a, string) ctxt xmlt2"
where
 [code]: "ctxt x = (
    XMLdo (STR ''box'') {xml_return Hole}
    XMLor XMLdo (STR ''funContext'') {
      name \<leftarrow> xml2name;
      left \<leftarrow> XMLdo (STR ''before'') {
        left \<leftarrow>* XTC_Parser.term xml2name;
        xml_return left
      };
      mid \<leftarrow> ctxt;
      right \<leftarrow> XMLdo (STR ''after'') {
        right \<leftarrow>* XTC_Parser.term xml2name;
        xml_return right
      };
      xml_return (More name left mid right)
   }) x"


definition flat_contexts :: "('a, string) ctxt list xmlt2" where "flat_contexts =
  xml_do (STR ''flatContexts'') (xml_take_many 0 \<infinity> (CPF_Parser.ctxt xml2name) xml_return)"

definition subst :: "('a, string) substL xmlt2"
  where
    "subst = XMLdo (STR ''substitution'') {
      list \<leftarrow>* XMLdo (STR ''substEntry'') {
        var \<leftarrow> xml_do (STR ''var'') (xml_take_text xml_return);
        trm \<leftarrow> XTC_Parser.term xml2name;
        xml_return (var,trm)
      };
      xml_return list
    }"

definition afs :: "'a afs_list xmlt2"
  where
    "afs = XMLdo (STR ''argumentFilter'') {
      ret \<leftarrow>* XMLdo (STR ''argumentFilterEntry'') {
        name \<leftarrow> xml2name;
        arity \<leftarrow> xml_nat (STR ''arity'');
        main \<leftarrow> xml_change (xml_nat (STR ''collapsing'')) (\<lambda>n. xml_return (Collapse (n - 1)))
          XMLor XMLdo (STR ''nonCollapsing'') {
            ls \<leftarrow>* XTC_Parser.position;
            xml_return (AFList ls)
          };
        xml_return ((name, arity), main)
      };
      xml_return ret
    }"

definition proj :: "'a projL xmlt2" where
  "proj = xml_change (CPF_Parser.afs xml2name) (\<lambda>afl. xml_return (Projection
    (map (\<lambda>(fa, e). case e of
      Collapse i \<Rightarrow> Pair fa i
    | _ \<Rightarrow> Pair fa (snd fa)) afl)))"


definition multiset_af :: "'a status_impl xmlt2" where
  "multiset_af = XMLdo (STR ''multisetArgumentFilter'') {
    ret \<leftarrow>* XMLdo (STR ''multisetArgumentFilterEntry'') {
      f \<leftarrow> xml2name;
      a \<leftarrow> xml_nat (STR ''arity'');
      p \<leftarrow> xml_do (STR ''status'') (xml_take_many 0 \<infinity> XTC_Parser.position xml_return);
      xml_return ((f,a), p)
    };
    xml_return ret
  }"

definition
  relstep :: "(pos \<times> ('a, string) rule \<times> bool \<times> ('a, string) term) xmlt2"
where
  "relstep = XMLdo (STR ''rewriteStep'') {
    p \<leftarrow> XTC_Parser.pos;
    r \<leftarrow> XTC_Parser.rule xml2name;
    rel \<leftarrow>? xml_leaf (STR ''relative'') ();
    t \<leftarrow> XTC_Parser.term xml2name;
    xml_return (p, r, rel = None, t)
  }"

definition
  rstep :: "(pos \<times> ('a, string) rule \<times> ('a, string) term) xmlt2"
where
  "rstep = XMLdo (STR ''rewriteStep'') {
    p \<leftarrow> XTC_Parser.pos;
    r \<leftarrow> XTC_Parser.rule xml2name;
    t \<leftarrow> XTC_Parser.term xml2name;
    xml_return (p, r, t)
  }"

definition
  estep :: "(pos \<times> ('a, string) rule \<times> bool \<times> ('a, string) term) xmlt2"
where
  "estep = XMLdo (STR ''equationStep'') {
    p \<leftarrow> XTC_Parser.pos;
    r \<leftarrow> XTC_Parser.rule xml2name;
    b \<leftarrow> xml_leaf (STR ''leftRight'') True XMLor xml_leaf (STR ''rightLeft'') False;
    t \<leftarrow> XTC_Parser.term xml2name;
    xml_return (p, r, b, t)
  }"

definition
  relsteps :: "(('a,string)term \<times> ('a,string)prseq) xmlt2"
where
  "relsteps = XMLdo (STR ''rewriteSequence'') {
    start \<leftarrow> xml_singleton (STR ''startTerm'') (XTC_Parser.term xml2name) id;
    steps \<leftarrow>* relstep;
    xml_return (start, steps)
  }"

definition
  rsteps :: "(('a, string) term \<times> ('a, string) rseq) xmlt2"
where
  "rsteps = XMLdo (STR ''rewriteSequence'') {
    start \<leftarrow> xml_singleton (STR ''startTerm'') (XTC_Parser.term xml2name) id;
    steps \<leftarrow>* CPF_Parser.rstep xml2name;
    xml_return (start, steps)
  }"


definition pat_eqv_prf :: "('a, string) pat_eqv_prf xmlt2"
  where
    "pat_eqv_prf =
      (let sub = CPF_Parser.subst xml2name in
      xml_singleton (STR ''patternEquivalence'') (
        XMLdo (STR ''domainRenaming'') {n \<leftarrow> sub; xml_return (Pat_Dom_Renaming n)}
        XMLor XMLdo (STR ''irrelevant'') {n \<leftarrow> sub; n2 \<leftarrow> sub; xml_return (Pat_Irrelevant n n2)}
        XMLor XMLdo (STR ''simplification'') {n \<leftarrow> sub; n2 \<leftarrow> sub; xml_return (Pat_Simplify n n2)}
      ) id)"

definition pat_term :: "(('a,string)term \<times> ('a,string)substL \<times> ('a,string)substL) xmlt2"
  where
    "pat_term = XMLdo (STR ''patternTerm'') {
      t \<leftarrow> XTC_Parser.term xml2name;
      s1 \<leftarrow> CPF_Parser.subst xml2name;
      s2 \<leftarrow> CPF_Parser.subst xml2name;
      xml_return (t,s1,s2)
    }"

abbreviation (input) pat_rule_fun where "pat_rule_fun x pat \<equiv> (let
   sub = CPF_Parser.subst xml2name;
   pt = pat_term;
   var = xml_text (STR ''var'')
in
(XMLdo (STR ''patternRule'') {
  pt1 \<leftarrow> pt;
  pt2 \<leftarrow> pt;
  z \<leftarrow> XMLdo (STR ''originalRule'') {
      r \<leftarrow> XTC_Parser.rule xml2name;
      b \<leftarrow> xml_bool (STR ''isPair'');
      xml_return (Pat_OrigRule r b)
    } XMLor XMLdo (STR ''initialPumping'') {
      a \<leftarrow> pat;
      b \<leftarrow> sub;
      c \<leftarrow> sub;
      xml_return (Pat_InitPump a b c)
    } XMLor XMLdo (STR ''initialPumpingContext'') {
      a \<leftarrow> pat;
      b \<leftarrow> sub;
      c \<leftarrow> XTC_Parser.pos;
      d \<leftarrow> var;
      xml_return (Pat_InitPumpCtxt a b c d)
    } XMLor XMLdo (STR ''equivalence'') {
      a \<leftarrow> pat;
      b \<leftarrow> xml_leaf (STR ''left'') True XMLor xml_leaf (STR ''right'') False;
      c \<leftarrow> CPF_Parser.pat_eqv_prf xml2name;
      xml_return (Pat_Equiv a b c)
    } XMLor XMLdo (STR ''narrowing'') {
      a \<leftarrow> pat;
      b \<leftarrow> pat;
      c \<leftarrow> XTC_Parser.pos;
      xml_return (Pat_Narrow a b c)
    } XMLor XMLdo (STR ''instantiation'') {
      a \<leftarrow> pat;
      b \<leftarrow> sub;
      c \<leftarrow> xml_leaf (STR ''base'') Pat_Base XMLor xml_leaf (STR ''pumping'') Pat_Pump XMLor xml_leaf (STR ''closing'') Pat_Close;
      xml_return (Pat_Inst a b c)
    } XMLor XMLdo (STR ''instantiationPumping'') {
      a \<leftarrow> pat;
      b \<leftarrow> xml_nat (STR ''power'');
      xml_return (Pat_Exp_Sigma a b)
    } XMLor XMLdo (STR ''rewriting'') {
      a \<leftarrow> pat;
      b \<leftarrow> CPF_Parser.rsteps xml2name;
      (po,va) \<leftarrow>
        xml_leaf (STR ''base'') (Pat_Base,[])
        XMLor XMLdo (STR ''pumping'') { v \<leftarrow> var; xml_return (Pat_Pump,v) }
        XMLor XMLdo (STR ''closing'') { v \<leftarrow> var; xml_return (Pat_Close,v) };
      xml_return (Pat_Rewr a b po va)
    };
   xml_return z
  }) x)"

partial_function (sum_bot) pat_rule_prf :: "('a,string)pat_rule_prf xmlt2"
  where "pat_rule_prf x = pat_rule_fun x pat_rule_prf"

(* one cannot put the 'let pat = ... ' directly into partial function, since then the monotonicity fails *)
lemma pat_rule_prf [code]:
  "pat_rule_prf x = (let pat = pat_rule_prf in pat_rule_fun x pat)"
     unfolding pat_rule_prf.simps[of x] Let_def by simp

definition
  conversion :: "(('a,string)term \<times> ('a,string) eseq) xmlt2"
where
  "conversion = XMLdo (STR ''conversion'') {
    a \<leftarrow> xml_singleton (STR ''startTerm'') (XTC_Parser.term xml2name) id;
    b \<leftarrow>* CPF_Parser.estep xml2name;
    xml_return (a,b)
  }"

definition subsumption_proof :: "('a,string)subsumption_proof xmlt2"
where "subsumption_proof = XMLdo (STR ''subsumptionProof'') {
  ret \<leftarrow>* XMLdo (STR ''ruleSubsumptionProof'') {
    r \<leftarrow> XTC_Parser.rule xml2name;
    (s,e) \<leftarrow> CPF_Parser.conversion xml2name;
    xml_return (r, e)
  };
  xml_return ret
}"

definition joinable_critical_pairs :: "('a,string)crit_pair_joins xmlt2"
  where
    "joinable_critical_pairs =
      (let rew = CPF_Parser.rsteps xml2name in
      XMLdo (STR ''joinableCriticalPairs'') {
        ret \<leftarrow>* XMLdo (STR ''joinableCriticalPair'') {
          (s,sseq) \<leftarrow> rew;
          (t,tseq) \<leftarrow> rew;
          xml_return (s,sseq,t,tseq)
        };
      xml_return ret
    })"

definition convertible_critical_peaks :: "('a,string) crit_peak_convs xmlt2"
where
  "convertible_critical_peaks =
    (let rew = CPF_Parser.rsteps xml2name in
    let conv = CPF_Parser.conversion xml2name in
    XMLdo (STR ''convertibleCriticalPeaks'') {
      ret \<leftarrow>* XMLdo (STR ''convertibleCriticalPeak'') {
        s \<leftarrow> xml_singleton (STR ''source'') (XTC_Parser.term xml2name) id;
        (s1,cl1,sl,cl2) \<leftarrow> XMLdo (STR ''conversionLeft'') {
          (s,c1) \<leftarrow> conv;
          (_,r) \<leftarrow> rew;
          (_,c2) \<leftarrow> conv;
          xml_return (s,c1,r,c2)
        };
        (s2,cr1,sr,cr2) \<leftarrow> XMLdo (STR ''conversionRight'') {
          (s,c1) \<leftarrow> conv;
          (_,r) \<leftarrow> rew;
          (_,c2) \<leftarrow> conv;
          xml_return (s,c1,r,c2)
        };
        xml_return (s,s1,cl1,sl,cl2,s2,cr1,sr,cr2)
      };
      xml_return ret
  })"

definition wcr_proof :: "'a join_info xmlt2"
  where "wcr_proof = xml_singleton (STR ''wcrProof'') (
    xml_change (joinable_critical_pairs) (xml_return \<circ> Guided)
    XMLor xml_change (xml_nat (STR ''joinableCriticalPairsBFS'')) (xml_return \<circ> Join_BFS)
    XMLor xml_leaf (STR ''joinableCriticalPairsAuto'') Join_NF
  ) id"

definition
  loop ::
    "(('a, string) term \<times> ('a, string) prseq \<times> ('a, string) substL \<times> ('a, string) ctxt) xmlt2"
where [code]:
  "loop = XMLdo (STR ''loop'') {
    (s,rseq) \<leftarrow> CPF_Parser.relsteps xml2name;
    \<sigma> \<leftarrow> CPF_Parser.subst xml2name;
    c \<leftarrow> CPF_Parser.ctxt xml2name;
    xml_return (s, rseq, \<sigma>, c)
  }"

definition
  start_term :: "('a, string) term \<Rightarrow> ('a, string) term xmlt2"
where
  "start_term t = XMLdo (STR ''startTerm'') {
    s \<leftarrow> XTC_Parser.term xml2name;
    if s = t then xml_return t
    else xml_error (STR ''<startTerm> does not match lhs'')
  }"

definition
  rseq ::
    "'a proj \<Rightarrow> ('a, string) rule \<Rightarrow> (('a,string)rule \<times> ('a,string)rseq)xmlt2"
where
  "rseq \<pi> r = XMLdo (STR ''rewriteSequence'') {
     _ \<leftarrow> CPF_Parser.start_term xml2name (proj_term \<pi> (fst r));
     rseq \<leftarrow>* CPF_Parser.rstep xml2name;
     xml_return (r, rseq)
  }"

definition
  projected_rseq ::
    "'a proj \<Rightarrow> (('a, string) rule \<times> ('a,string)rseq) xmlt2"
where
  "projected_rseq \<pi> =
    XMLdo (STR ''projectedRewriteSequence'') {
      r \<leftarrow> XTC_Parser.rule xml2name;
      ret \<leftarrow> CPF_Parser.rseq xml2name \<pi> r;
      xml_return ret
    }"



definition status_precedence :: "('a status_prec_repr) xmlt2" where
  "status_precedence = XMLdo (STR ''statusPrecedence'') {
    ret \<leftarrow>* XMLdo (STR ''statusPrecedenceEntry'') {
      f \<leftarrow> xml2name;
      a \<leftarrow> xml_nat (STR ''arity'');
      p \<leftarrow> xml_nat (STR ''precedence'');
      s \<leftarrow> xml_leaf (STR ''lex'') Lex XMLor xml_leaf (STR ''mul'') Mul;
      xml_return ((f,a), (p,s))
    };
    xml_return ret
  }"

definition path_order :: "'a redtriple_impl xmlt2"
  where
    "path_order = XMLdo (STR ''pathOrder'') {
      prec\<tau> \<leftarrow> CPF_Parser.status_precedence xml2name;
      af \<leftarrow>[[]] CPF_Parser.afs xml2name;
      xml_return (RPO prec\<tau> af)
    }"

definition precedence_weight :: "(nat \<Rightarrow> 'a prec_weight_repr) xmlt2"
  where
    "precedence_weight = XMLdo (STR ''precedenceWeight'') {
      ret \<leftarrow>* XMLdo (STR ''precedenceWeightEntry'') {
        f \<leftarrow> xml2name;
        a \<leftarrow> xml_nat (STR ''arity'');
        p \<leftarrow> xml_nat (STR ''precedence'');
        w \<leftarrow> xml_nat (STR ''weight'');
        e \<leftarrow>? XMLdo (STR ''subtermCoefficientEntries'') {
          ret \<leftarrow>* xml_nat (STR ''entry'');
          xml_return ret
        };
        xml_return ((f,a), (p,w,e))
      };
      xml_return (Pair ret)
    }"


definition precedence_weight_ac :: "(nat \<Rightarrow> 'a prec_weight_ac_repr) xmlt2" where
  "precedence_weight_ac = XMLdo (STR ''precedenceWeight'') {
    ret \<leftarrow>* XMLdo (STR ''precedenceWeightEntry'') {
      f \<leftarrow> xml2name;
      a \<leftarrow> xml_nat (STR ''arity'');
      p \<leftarrow> xml_nat (STR ''precedence'');
      w \<leftarrow> xml_nat (STR ''weight'');
      e \<leftarrow>[False] xml_bool (STR ''isAC'');
      xml_return ((f,a), (p,w,e))
    };
    xml_return (Pair ret)
  }"


definition wpo_params :: "'a wpo_params xmlt2" where
  "wpo_params = XMLdo (STR ''precedenceStatus'') {
    ret \<leftarrow>* XMLdo (STR ''precedenceStatusEntry'') {
      f \<leftarrow> xml2name;
      a \<leftarrow> xml_nat (STR ''arity'');
      p \<leftarrow> xml_nat (STR ''precedence'');
      s \<leftarrow> XMLdo (STR ''status'') {ps \<leftarrow>* XTC_Parser.position; xml_return ps};
      xml_return ((f,a), (p,s))
    };
    xml_return ret
  }"

definition knuth_bendix_order :: "'a redtriple_impl xmlt2" where
  "knuth_bendix_order = XMLdo (STR ''knuthBendixOrder'') {
   w0 \<leftarrow> xml_nat (STR ''w0'');
   prw \<leftarrow> CPF_Parser.precedence_weight xml2name;
   afs \<leftarrow>[[]] CPF_Parser.afs xml2name;
   xml_return (KBO (prw w0) afs)
  }"

definition ac_knuth_bendix_order :: "'a redtriple_impl xmlt2" where
  "ac_knuth_bendix_order = XMLdo (STR ''ACKBO'') {
   w0 \<leftarrow> xml_nat (STR ''w0'');
   prw \<leftarrow> CPF_Parser.precedence_weight_ac xml2name;
   afs \<leftarrow>[[]] CPF_Parser.afs xml2name;
   xml_return (ACKBO (prw w0) afs)
  }"

partial_function (sum_bot) exp_parser :: "(max_poly.sig, nat) term xmlt2" where
  [code]: "exp_parser xml = (
  XMLdo (STR ''product'') {
    exps \<leftarrow>* exp_parser;
    xml_return (Fun max_poly.ProdF exps)
  } XMLor XMLdo (STR ''sum'') {
    exps \<leftarrow>* exp_parser;
    xml_return (Fun max_poly.SumF exps)
  } XMLor XMLdo (STR ''max'') {
    exps \<leftarrow>^{1..\<infinity>} exp_parser;
    xml_return (Fun max_poly.MaxF exps)
  } XMLor XMLdo (STR ''constant'') {n \<leftarrow>nat; xml_return (max_poly.const n)}
    XMLor XMLdo (STR ''variable'') {n \<leftarrow>nat; xml_return (Var (n - 1))}
  ) xml"

partial_function (sum_bot) redtriple :: "bool \<Rightarrow> 'a redtriple_impl xmlt2" where
  [code]: "redtriple bi xml = (
    path_order
    XMLor knuth_bendix_order
    XMLor ac_knuth_bendix_order
    XMLor CPF_Parser.interpretation bi xml2name
    XMLor XMLdo (STR ''maxPoly'') {
      inters \<leftarrow>* XMLdo (STR ''interpret'') {
        f \<leftarrow> xml2name;
        a \<leftarrow> xml_nat (STR ''arity'');
        e \<leftarrow> exp_parser;
        xml_return ((f, a), e)
      };
      xml_return (Max_poly inters)
    } XMLor XMLdo (STR ''weightedPathOrder'') {
      a \<leftarrow> wpo_params;
      b \<leftarrow> redtriple bi;
      xml_return (WPO a b)
    }
    XMLor XMLdo (STR ''filteredRedPair'') {
      af \<leftarrow>[[]] CPF_Parser.afs xml2name;
      b \<leftarrow> redtriple bi;
      xml_return (Filtered_Redtriple af b)
    }
   ) xml"

definition level_mapping :: "'a scnp_af xmlt2" where
  "level_mapping = XMLdo (STR ''levelMapping'') {
    ret \<leftarrow>* XMLdo (STR ''levelMappingEntry'') {
      f \<leftarrow> xml2name;
      a \<leftarrow> xml_nat (STR ''arity'');
      ps \<leftarrow>* xml_pair (STR ''positionLevelEntry'')
         (xml_nat (STR ''position''))
         (xml_nat (STR ''level''))
         Pair;
      xml_return ((f,a), map (\<lambda> (p,l). (if p = 0 then a else p - 1,l)) ps)
    };
    xml_return ret
  }"

definition redtriple' where
  "redtriple' bi = XMLdo (STR ''redPair'') {
    ret \<leftarrow> XMLdo (STR ''scnp'') {
      a \<leftarrow> XMLdo (STR ''status'') {
        ret \<leftarrow>
          xml_leaf (STR ''ms'') MS_Ext
          XMLor xml_leaf (STR ''min'') Min_Ext
          XMLor xml_leaf (STR ''dms'') Dms_Ext
          XMLor xml_leaf (STR ''max'') Max_Ext;
        xml_return ret
      };
      b \<leftarrow> level_mapping;
      c \<leftarrow> xml_singleton (STR ''redPair'') (redtriple False) id;
      xml_return (Inl (SCNP a b c))
    } XMLor xml_change (redtriple bi) (xml_return \<circ> Inr);
  xml_return ret
}"


definition
  ordering_constraint_proof :: "bool \<Rightarrow> 'a redtriple_impl xmlt2"
where
  "ordering_constraint_proof bi =
    xml_singleton (STR ''orderingConstraintProof'') (
      xml_singleton (STR ''redPair'') (redtriple bi) id
    ) id"

definition
  ordering_constraint_proof'
where
  "ordering_constraint_proof' bi =
     xml_singleton (STR ''orderingConstraintProof'') (redtriple' bi) id"

definition
  id_labeling :: "(('f,'a list)lab,'a,'a list + ('f,'a list)lab list)label"
where
  "id_labeling _ cs = Inl cs"

definition scg_position :: "nat xmlt2" where
  "scg_position = xml_nat (STR ''position'')"

definition
  scg :: "(('a, string) rule \<times> ((nat \<times> nat) list \<times> (nat \<times> nat) list)) xmlt2"
where
  "scg = XMLdo (STR ''sizeChangeGraph'') {
    lr \<leftarrow> XTC_Parser.rule xml2name;
    edges \<leftarrow>* XMLdo (STR ''edge'') {
       p \<leftarrow> CPF_Parser.scg_position;
       s \<leftarrow> xml_bool (STR ''strict'');
       q \<leftarrow> CPF_Parser.scg_position;
       xml_return (s,(p,q))
    };
    xml_return (lr,(map snd (filter fst edges), map snd (filter (\<lambda> x. \<not> fst x) edges)))
  }"

definition uncurry_info :: "('a, string) uncurry_info xmlt2" where
  "uncurry_info = XMLdo (STR ''uncurryInformation'') {
    a \<leftarrow> xml2name;
    sml \<leftarrow> XMLdo (STR ''uncurriedSymbols'') {
      ret \<leftarrow>* XMLdo (STR ''uncurriedSymbolEntry'') {
        f \<leftarrow> xml2name;
        n \<leftarrow> xml_nat (STR ''arity'');
        fs \<leftarrow>* xml2name;
        xml_return ((f, n), fs)
      };
      xml_return ret
    };
    U \<leftarrow> xml_singleton (STR ''uncurryRules'') (XTC_Parser.rules xml2name) id;
    E \<leftarrow> xml_singleton (STR ''etaRules'') (XTC_Parser.rules xml2name) id;
    xml_return (a, sml, U, E)
  }"


definition xml2dp_input
where
  "xml2dp_input = XMLdo (STR ''dpInput'') {
    r \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
    p \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
    s \<leftarrow>[No_Strategy] strategy xml2name;
    m \<leftarrow> xml_bool (STR ''minimal'');
    xml_return (DP_input m p s r)
  }"

partial_function (sum_bot)
  xml2cond_constraint :: "('a,string) cond_constraint xmlt2" where
  [code]: "xml2cond_constraint x = xml_singleton (STR ''conditionalConstraint'') (
    XMLdo (STR ''all'') {
      a \<leftarrow> xml_text (STR ''var'');
      b \<leftarrow> xml2cond_constraint;
      xml_return (CC_all a b)
    } XMLor XMLdo (STR ''implication'') {
      c \<leftarrow> xml2cond_constraint;
      cs \<leftarrow>* xml2cond_constraint;
      let ccs = c # cs;
      xml_return (CC_impl (take (length cs) ccs) (last ccs))
    } XMLor XMLdo (STR ''constraint'') {
      s \<leftarrow> XTC_Parser.term xml2name;
      rel \<leftarrow>
        xml_leaf (STR ''rewrite'') None
        XMLor xml_leaf (STR ''strict'') (Some True)
        XMLor xml_leaf (STR ''nonStrict'') (Some False);
      t \<leftarrow> XTC_Parser.term xml2name;
      xml_return (case rel of None \<Rightarrow> CC_rewr s t | Some stri \<Rightarrow> CC_cond stri (s,t))
    }) id x"


partial_function (sum_bot) xml2cond_constraint_prf :: "('a,string) cond_constraint_prf xmlt2" where
  [code]: "xml2cond_constraint_prf x = (
  let cc = xml2cond_constraint in
  xml_singleton (STR ''conditionalConstraintProof'') (
    xml_leaf (STR ''final'') Final
    XMLor XMLdo (STR ''differentConstructor'') {
      a \<leftarrow> cc;
      xml_return (Different_Constructor a)
    } XMLor XMLdo (STR ''sameConstructor'') {
      a \<leftarrow> cc; b \<leftarrow> cc; c \<leftarrow> xml2cond_constraint_prf;
      xml_return (Same_Constructor a b c)
    } XMLor XMLdo (STR ''variableEquation'') {
      a \<leftarrow> xml_text (STR ''var''); b \<leftarrow> XTC_Parser.term xml2name; c \<leftarrow>  cc;
      d \<leftarrow> xml2cond_constraint_prf;
      xml_return (Variable_Equation a b c d)
    } XMLor XMLdo (STR ''funargIntoVar'') {
      a \<leftarrow> cc;
      b \<leftarrow> XTC_Parser.position;
      c \<leftarrow> xml_text (STR ''var'');
      d \<leftarrow> cc;
      e \<leftarrow> (xml2cond_constraint_prf);
      xml_return (Funarg_Into_Var a b c d e)
    } XMLor XMLdo (STR ''simplifyCondition'') {
      a \<leftarrow> cc; b \<leftarrow> (CPF_Parser.subst xml2name); c \<leftarrow> cc; d \<leftarrow> (xml2cond_constraint_prf);
      xml_return (Simplify_Condition a b c d)
    } XMLor XMLdo (STR ''induction'') {
      a \<leftarrow> cc;
      b \<leftarrow> XMLdo (STR ''conjuncts'') {ret \<leftarrow>* cc; xml_return ret};
      c \<leftarrow> XMLdo (STR ''ruleConstraintProofs'') {
        ret \<leftarrow>* XMLdo (STR ''ruleConstraintProof'') {
          lr \<leftarrow> XTC_Parser.rule xml2name;
          rys \<leftarrow> XMLdo (STR ''subtermVarEntries'') {
            ret \<leftarrow>* XMLdo (STR ''subtermVarEntry'') {
              a \<leftarrow> XTC_Parser.term xml2name;
              b \<leftarrow>* xml_text (STR ''var'');
              xml_return (a,b)
            };
            xml_return ret
          };
          cc \<leftarrow> cc;
          p \<leftarrow> xml2cond_constraint_prf;
          xml_return (lr,rys,cc,p)
        };
        xml_return ret
      };
      xml_return (Induction a b c)
    } XMLor XMLdo (STR ''deleteCondition'') {
      a \<leftarrow> cc;
      b \<leftarrow> xml2cond_constraint_prf;
      xml_return (Delete_Condition a b)
   }) id
    x)"

definition
  xml2cond_red_pair_proof :: "('a, string) cond_red_pair_prf xmlt2" where
  "xml2cond_red_pair_proof = XMLdo (STR ''condRedPairProof'') {
    c \<leftarrow> xml2name;
    b \<leftarrow> xml_nat (STR ''before'');
    a \<leftarrow> xml_nat (STR ''after'');
    ccs \<leftarrow> XMLdo (STR ''conditions'') {
      ret \<leftarrow>* XMLdo (STR ''condition'') {
         c \<leftarrow> xml2cond_constraint;
         s \<leftarrow> xml_singleton (STR ''dpSequence'') (XTC_Parser.rules xml2name) id;
         p \<leftarrow> xml2cond_constraint_prf;
         xml_return (c,s,p)
      };
      xml_return ret
    };
    xml_return (Cond_Red_Pair_Prf c ccs b a)
  }"

definition renaming :: "('a \<times> 'a)list xmlt2" where
  "renaming = XMLdo (STR ''renaming'') {
    ret \<leftarrow>* XMLdo (STR ''renamingEntry'') {
      a \<leftarrow> xml2name;
      b \<leftarrow> xml2name;
      xml_return (a,b)
    };
    xml_return ret
  }"

(* TODO: improve CPF and remove this *)
definition xml2complexity_input
where "xml2complexity_input = XMLdo (STR ''complexityInput'') {
    (q, r, s, start) \<leftarrow> xml_change (xml2_trs_input xml2name) (\<lambda>inp.
      case inp of Inn_TRS_input q r s start \<Rightarrow> xml_return (q, r, s, start)
      |  _ \<Rightarrow> xml_error (STR ''require innermost trs at this point'')
    );
    cm \<leftarrow> complexity_measure xml2name;
    cc \<leftarrow> complexity_class;
    xml_return (CPX_input q r s cm cc)
  }"

definition symbols_no_arity :: "ltag \<Rightarrow> 'a list xmlt2"
where
  "symbols_no_arity tagname =
    XMLdo tagname {ret \<leftarrow>* xml2name; xml_return ret}"

definition
  xml2ac_tp_input :: "('a,string) input xmlt2"
where
  "xml2ac_tp_input = XMLdo (STR ''acRewriteSystem'') {
     a \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
     b \<leftarrow> symbols_no_arity (STR ''Asymbols'');
     c \<leftarrow> symbols_no_arity (STR ''Csymbols'');
     xml_return (AC_input a b c)
  }"

definition
  xml2_commutation_input :: "('a,string) input xmlt2"
where
  "xml2_commutation_input = XMLdo (STR ''commutationInput'') {
     f \<leftarrow> CPFsignature xml2name;
     r \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
     s \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
     xml_return (Commutation_input f r s)
  }"

definition
  xml2input :: "('a,string) input xmlt2"
where
  "xml2input = xml_singleton (STR ''input'') (
    xml2dp_input
    XMLor xml2completion_input xml2name
    XMLor xml2ordered_completion_input xml2name
    XMLor xml2_trs_input xml2name
    XMLor xml2_commutation_input 
    XMLor xml2equational_input xml2name
    XMLor xml2complexity_input
    XMLor xml2ctrs_input xml2name
    XMLor XMLdo (STR ''treeAutomatonProblem'') {
      a \<leftarrow> tree_automaton (ta_normal_lhs xml2name);
      b \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      xml_return (TA_input a b)
    } XMLor xml2ac_tp_input
    XMLor xml_change lts_input_parser (xml_return \<circ> LTS_input)
    XMLor xml_change lts_safety_input_parser (\<lambda>(x,y). xml_return (LTS_safety_input x y))
    XMLor xml2infeasibility_input xml2name
    XMLor xml_change xml2unknown_input (xml_return \<circ> Unknown_input)
  ) id"


text \<open>
  Parse a list of pairs of rules, where the first argument gives the name of the outermost tag
  and the second argument the name of the tag around each pair.
\<close>
definition
  rule_pairs :: "ltag \<Rightarrow> ltag \<Rightarrow> (('a, string) rule \<times> ('a, string) rule) list xmlt2"
where
  "rule_pairs s p = XMLdo s {
    ret \<leftarrow>* xml_pair p
      (XTC_Parser.rule xml2name)
      (XTC_Parser.rule xml2name)
      Pair;
    xml_return ret
  }"

definition xml2inn_fp_trs_assm where
  "xml2inn_fp_trs_assm = xml_change (xml2_trs_input xml2name) (\<lambda>inp.
    case inp of
      Inn_TRS_input inn r s start \<Rightarrow>
        if start \<noteq> start_term.Full then
          xml_error (STR ''start term is not allowed here'')
        else
          xml_return (Inl (default_nfs_trs, strategy_to_Q inn r, r, s))
    | FP_TRS_input fp r \<Rightarrow> xml_return (Inr (strategy_to_fp fp r, r))
    | _ \<Rightarrow> xml_error (STR ''trs as input required'')
  )"

definition xml2inn_trs_assm where
  "xml2inn_trs_assm = xml_change xml2inn_fp_trs_assm (\<lambda>inp.
    case inp of
      Inl i \<Rightarrow> xml_return i
    | _ \<Rightarrow> xml_error (STR ''innermost (relative) TRS expected at this point'')
   )"

definition xml2fp_trs_assm where
  "xml2fp_trs_assm = xml_change xml2inn_fp_trs_assm (\<lambda>inp.
    case inp of
      Inr i \<Rightarrow> xml_return i
    | _ \<Rightarrow> xml_error (STR ''FP TRS expected at this point'')
   )"

definition xml2dp_input' where
  "xml2dp_input' termination = xml_change xml2dp_input
    (\<lambda>ret. case ret of DP_input m p q r \<Rightarrow>
     xml_return (if termination then default_nfs_dp else default_nfs_nt_dp, m, p, [], strategy_to_Q q r, [], r))"

definition xml2complexity_input' where
  "xml2complexity_input' = xml_change xml2complexity_input
    (\<lambda> ret. case ret of CPX_input q s w cm cc \<Rightarrow> xml_return (strategy_to_Q q (s @ w), s, w, cm,cc))"
end

context fixes xml2name :: "('f :: {showl,compare_order}, 'l :: {showl,compare_order}) lab xmlt2"
begin
partial_function_mr (sum_bot)
  xml2dp_termination_proof :: "('f, 'l, string) dp_termination_proof xmlt2"
  and xml2trs_termination_proof_inner
  and xml2trs_termination_proof :: "('f, 'l, string) trs_termination_proof xmlt2"
  and xml2fptrs_termination_proof :: "('f, 'l, string) fptrs_termination_proof xmlt2"
  and xml2unknown_proof :: "('f, 'l, string) unknown_proof xmlt2"
  and xml2subproof
where
  "xml2dp_termination_proof x =
  xml_singleton (STR ''dpProof'') (
    XMLdo (STR ''pIsEmpty'') {
      xml_return P_is_Empty
    } XMLor XMLdo (STR ''depGraphProc'') {
      a \<leftarrow>* XMLdo (STR ''component'') {
        dps \<leftarrow> (xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id);
        _ \<leftarrow>? xml_bool (STR ''realScc''); \<comment> \<open>this is redundant, and the original syntax is not expressible by XML scheme\<close>
        _ \<leftarrow>? XMLdo (STR ''arcs'') { _ \<leftarrow>* xml_any; xml_return () };
        prfOpt \<leftarrow>? xml2dp_termination_proof;
        xml_return (prfOpt, dps)
      };
      xml_return (Dep_Graph_Proc a)
    } XMLor XMLdo (STR ''redPairProc'') {
      a \<leftarrow> CPF_Parser.ordering_constraint_proof' xml2name False;
      b \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
      c \<leftarrow> xml2dp_termination_proof;
      xml_return (Redpair_Proc a b c)
    } XMLor XMLdo (STR ''usableRulesProc'') {
      a \<leftarrow> xml_singleton (STR ''usableRules'') (XTC_Parser.rules xml2name) id;
      b \<leftarrow> xml2dp_termination_proof;
      xml_return (Usable_Rules_Proc a b)
    } XMLor XMLdo (STR ''innermostLhssRemovalProc'') {
      a \<leftarrow> innermostLhss xml2name;
      b \<leftarrow> xml2dp_termination_proof;
      xml_return (Q_Reduction_Proc a b)
    } XMLor XMLdo (STR ''rewritingProc'') {
      (s,t) \<leftarrow> XTC_Parser.rule xml2name;
      (p,lr,t') \<leftarrow> CPF_Parser.rstep xml2name;
      st'' \<leftarrow>[(s,t')] XTC_Parser.rule xml2name;
      ur \<leftarrow>? xml_singleton (STR ''usableRules'') (XTC_Parser.rules xml2name) id;
      prof \<leftarrow> xml2dp_termination_proof;
      xml_return (Rewriting_Proc ur (s,t) (s,t') st'' lr p prof)
    } XMLor XMLdo (STR ''narrowingProc'') {
      a \<leftarrow> XTC_Parser.rule xml2name;
      b \<leftarrow> XTC_Parser.pos;
      c \<leftarrow> xml_singleton (STR ''narrowings'') (XTC_Parser.rules xml2name) id;
      d \<leftarrow> xml2dp_termination_proof;
      xml_return (Narrowing_Proc a b c d)
    } XMLor XMLdo (STR ''instantiationProc'') {
      a \<leftarrow> XTC_Parser.rule xml2name;
      b \<leftarrow> xml_singleton (STR ''instantiations'') (XTC_Parser.rules xml2name) id;
      c \<leftarrow> xml2dp_termination_proof;
      xml_return (Instantiation_Proc a b c)
    } XMLor XMLdo (STR ''forwardInstantiationProc'') {
      a \<leftarrow> XTC_Parser.rule xml2name;
      b \<leftarrow> xml_singleton (STR ''instantiations'') (XTC_Parser.rules xml2name) id;
      c \<leftarrow>? xml_singleton (STR ''usableRules'') (XTC_Parser.rules xml2name) id;
      d \<leftarrow> xml2dp_termination_proof;
      xml_return (Forward_Instantiation_Proc a b c d)
    } XMLor XMLdo (STR ''semlabProc'') {
      sli \<leftarrow> CPF_Parser.sl_variant xml2name;
      lp \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
      lr \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      lq \<leftarrow>[[]] innermostLhss xml2name;
      p \<leftarrow> xml2dp_termination_proof;
      xml_return (Semlab_Proc sli lp lq lr p)
    } XMLor XMLdo (STR ''subtermProc'') {
      pi_mpi \<leftarrow> xml_change (CPF_Parser.proj xml2name) (xml_return \<circ> Inl)
        XMLor xml_change (CPF_Parser.multiset_af xml2name) (xml_return \<circ> Inr);
      seq \<leftarrow>* CPF_Parser.projected_rseq xml2name (create_proj (case pi_mpi of Inl lpi \<Rightarrow> lpi | Inr mpi \<Rightarrow> Projection []));
      dps \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
      prf \<leftarrow> xml2dp_termination_proof;
      xml_return (case pi_mpi of
         Inl lpi \<Rightarrow> Subterm_Criterion_Proc lpi seq dps prf
       | Inr mpi \<Rightarrow> Gen_Subterm_Criterion_Proc mpi dps prf)
    } XMLor XMLdo (STR ''redPairUrProc'') {
      a \<leftarrow> CPF_Parser.ordering_constraint_proof' xml2name False;
      b \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
      c \<leftarrow> xml_singleton (STR ''usableRules'') (XTC_Parser.rules xml2name) id;
      d \<leftarrow> xml2dp_termination_proof;
      xml_return (Redpair_UR_Proc a b c d)
    } XMLor XMLdo (STR ''monoRedPairUrProc'') {
      a \<leftarrow> CPF_Parser.ordering_constraint_proof xml2name False;
      b \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
      c \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      d \<leftarrow> xml_singleton (STR ''usableRules'') (XTC_Parser.rules xml2name) id;
      e \<leftarrow> xml2dp_termination_proof;
      xml_return (Mono_Redpair_UR_Proc a b c d e)
    } XMLor XMLdo (STR ''monoRedPairProc'') {
      a \<leftarrow> CPF_Parser.ordering_constraint_proof xml2name False;
      b \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
      c \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      d \<leftarrow> xml2dp_termination_proof;
      xml_return (Mono_Redpair_Proc a b c d)
    } XMLor XMLdo (STR ''innermostMonoRedPairProc'') {
      rp \<leftarrow> CPF_Parser.ordering_constraint_proof xml2name False;
      (p,r) \<leftarrow> XMLdo (STR ''deleted'') {
        x \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
        y \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
        xml_return (x,y)
      };
      c \<leftarrow> xml2dp_termination_proof;
      xml_return (Mono_URM_Redpair_Proc rp p r c)
    } XMLor XMLdo (STR ''uncurryProc'') {
      a \<leftarrow>? xml_nat (STR ''applicativeTop'');
      b \<leftarrow> CPF_Parser.uncurry_info xml2name;
      c \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
      d \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      e \<leftarrow> xml2dp_termination_proof;
      xml_return (Uncurry_Proc a b c d e)
    } XMLor XMLdo (STR ''flatContextClosureProc'') {
      a \<leftarrow> xml_singleton (STR ''freshSymbol'') xml2name id;
      b \<leftarrow> CPF_Parser.flat_contexts xml2name;
      c \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
      d \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      e \<leftarrow> xml2dp_termination_proof;
      xml_return (Fcc_Proc a b c d e)
    } XMLor XMLdo (STR ''switchInnermostProc'') {
      a \<leftarrow> CPF_Parser.wcr_proof xml2name;
      b \<leftarrow> xml2dp_termination_proof;
      xml_return (Switch_Innermost_Proc a b)
    } XMLor XMLdo (STR ''splitProc'') {
      a \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
      b \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      c \<leftarrow> xml2dp_termination_proof;
      d \<leftarrow> xml2dp_termination_proof;
      xml_return (Split_Proc a b c d)
    } XMLor XMLdo (STR ''finitenessAssumption'') {
      dpp \<leftarrow> xml2dp_input' xml2name True;
      xml_return (Assume_Finite dpp [])
    } XMLor XMLdo (STR ''unknownProof'') {
      a \<leftarrow> xml_text (STR ''description'');
      b \<leftarrow> xml2dp_input' xml2name True;
      c \<leftarrow>* xml2subproof;
      xml_return (Assume_Finite b c)
    } XMLor XMLdo (STR ''switchToTRS'') {
      a \<leftarrow> xml2trs_termination_proof;
      xml_return (To_Trs_Proc a)
    } XMLor XMLdo (STR ''unlabProc'') {
      a \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
      b \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      c \<leftarrow> xml2dp_termination_proof;
      xml_return (Unlab_Proc a b c)
    } XMLor XMLdo (STR ''generalRedPairProc'') {
      rp \<leftarrow> CPF_Parser.ordering_constraint_proof xml2name True;
      s \<leftarrow> xml_singleton (STR ''strict'') (XTC_Parser.rules xml2name) id;
      b \<leftarrow> xml_singleton (STR ''bound'') (XTC_Parser.rules xml2name) id;
      c \<leftarrow> CPF_Parser.xml2cond_red_pair_proof xml2name;
      ps \<leftarrow> xml2dp_termination_proof;
      pbo \<leftarrow>? xml2dp_termination_proof;
      xml_return (General_Redpair_Proc rp s b c (case pbo of None \<Rightarrow> [ps] | Some pb \<Rightarrow> [ps,pb]))
    } XMLor XMLdo (STR ''complexConstantRemovalProc'') {
      t \<leftarrow> XTC_Parser.term xml2name;
      rls \<leftarrow> XMLdo (STR ''ruleMap'') {
        ret \<leftarrow>* XMLdo (STR ''ruleMapEntry'') {
          a \<leftarrow> XTC_Parser.rule xml2name;
          b \<leftarrow> XTC_Parser.rule xml2name;
          xml_return (a,b)
        };
        xml_return ret
      };
      prf \<leftarrow> xml2dp_termination_proof;
      xml_return (Complex_Constant_Removal_Proc (Complex_Constant_Removal_Proof t rls) prf)
    } XMLor XMLdo (STR ''sizeChangeProc'') {
      version \<leftarrow> xml_leaf (STR ''subtermCriterion'') None
        XMLor XMLdo (STR ''reductionPair'') {
          redp \<leftarrow> CPF_Parser.ordering_constraint_proof xml2name False;
          ur \<leftarrow>? xml_singleton (STR ''usableRules'') (XTC_Parser.rules xml2name) id;
          xml_return (Some (redp,ur))
        };
      b \<leftarrow>* CPF_Parser.scg xml2name;
      case version of
          None \<Rightarrow> xml_return (Size_Change_Subterm_Proc b)
        | Some redp_ur \<Rightarrow> xml_return (Size_Change_Redpair_Proc (fst redp_ur) (snd redp_ur) b)
    }
  ) id x"
| "xml2trs_termination_proof_inner x = (
    XMLdo (STR ''rIsEmpty'') {
      xml_return R_is_Empty
    } XMLor XMLdo (STR ''sIsEmpty'') {
      p \<leftarrow> xml2trs_termination_proof;
      xml_return p
    } XMLor XMLdo (STR ''semlab'') {
       sli \<leftarrow> CPF_Parser.sl_variant xml2name;
       lr \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
       lrw \<leftarrow>[[]] xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
       lq \<leftarrow>[[]] innermostLhss xml2name;
       p \<leftarrow> xml2trs_termination_proof;
       xml_return (Semlab sli lq (lr@lrw) p)
    } XMLor XMLdo (STR ''split'') {
      a \<leftarrow> (xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id);
      b \<leftarrow> xml2trs_termination_proof;
      c \<leftarrow> xml2trs_termination_proof;
      xml_return (Split a b c)
    } XMLor XMLdo (STR ''dpTrans'') {
      dps \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
      _ \<leftarrow> xml_bool (STR ''markedSymbols'');
      prf \<leftarrow> xml2dp_termination_proof;
      xml_return (DP_Trans default_nfs_dp True dps prf)
    } XMLor XMLdo (STR ''ruleRemoval'') {
      ord \<leftarrow> CPF_Parser.ordering_constraint_proof xml2name False;
      r \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      s \<leftarrow>[[]] xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      p \<leftarrow> xml2trs_termination_proof;
      xml_return (Rule_Removal ord (r@s) p)
    } XMLor xml_change (bounds_info xml2name) (xml_return \<circ> Bounds)
    XMLor XMLdo (STR ''stringReversal'') {
      _ \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      _ \<leftarrow>? xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      prf \<leftarrow> xml2trs_termination_proof;
      xml_return (String_Reversal prf)
    } XMLor XMLdo (STR ''equalityRemoval'') {
      a \<leftarrow> xml2trs_termination_proof;
      xml_return (Drop_Equality a)
    } XMLor XMLdo (STR ''rightGroundTermination'') {
      xml_return Right_Ground_Termination
    } XMLor XMLdo (STR ''constantToUnary'') {
       v \<leftarrow> XTC_Parser.plain_var;
       ren \<leftarrow> CPF_Parser.renaming xml2name;
       S \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
       Sw \<leftarrow>[[]] xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
       p \<leftarrow> xml2trs_termination_proof;
       xml_return (Constant_String (Const_string_sound_proof v ren S Sw) p)
    } XMLor XMLdo (STR ''removeNonApplicableRules'') {
      a \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      b \<leftarrow> xml2trs_termination_proof;
      xml_return (Remove_Nonapplicable_Rules a b)
    } XMLor XMLdo (STR ''uncurry'') {
      i \<leftarrow> CPF_Parser.uncurry_info xml2name;
      r \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      s \<leftarrow>[[]] xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      p \<leftarrow> xml2trs_termination_proof;
      xml_return (Uncurry i (r@s) p)
    } XMLor XMLdo (STR ''flatContextClosure'') {
      i \<leftarrow> CPF_Parser.flat_contexts xml2name;
      r \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      s \<leftarrow>[[]] xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      p \<leftarrow> xml2trs_termination_proof;
      xml_return (Fcc i (r@s) p)
    } XMLor XMLdo (STR ''switchInnermost'') {
      a \<leftarrow> CPF_Parser.wcr_proof xml2name;
      b \<leftarrow> xml2trs_termination_proof;
      xml_return (Switch_Innermost a b)
    } XMLor XMLdo (STR ''permutingArgumentFilter'') {
      a \<leftarrow> CPF_Parser.afs xml2name;
      b \<leftarrow> xml2trs_termination_proof;
      xml_return (Permuting_AFS a b)
    } XMLor XMLdo (STR ''terminationAssumption'') {
      qtrs \<leftarrow> xml2inn_trs_assm xml2name;
      xml_return (Assume_SN qtrs [])
    } XMLor XMLdo (STR ''relativeTerminationAssumption'') {
       ret \<leftarrow> XMLdo (STR ''trsInput'') {
         r \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
         rw \<leftarrow>[[]] xml_singleton (STR ''relativeRules'') (XTC_Parser.rules xml2name) id;
         xml_return (Assume_SN (default_nfs_trs, [], r, rw) [])
      };
      xml_return ret
    } XMLor XMLdo (STR ''unknownProof'') {
        a \<leftarrow> xml_text (STR ''description'');
        b \<leftarrow> xml2inn_trs_assm xml2name;
        c \<leftarrow>* xml2subproof;
        xml_return (Assume_SN b c)
    }
  ) x"
| "xml2trs_termination_proof x = (
    xml_singleton (STR ''trsTerminationProof'') xml2trs_termination_proof_inner id
    XMLor xml_singleton (STR ''relativeTerminationProof'') xml2trs_termination_proof_inner id
  ) x"
| "xml2fptrs_termination_proof x = (
    XMLdo (STR ''trsTerminationProof'') {
      r \<leftarrow> XMLdo (STR ''terminationAssumption'') {
        inp \<leftarrow> xml2fp_trs_assm xml2name;
        xml_return (Assume_FP_SN inp [])
      } XMLor XMLdo (STR ''unknownProof'') {
        inp \<leftarrow> xml2fp_trs_assm xml2name;
        subprfs \<leftarrow>* xml2subproof;
        xml_return (Assume_FP_SN inp subprfs)
      };
      xml_return r
    }
  ) x"
| "xml2unknown_proof x =
  XMLdo (STR ''unknownInputProof'') {
    ret \<leftarrow> XMLdo (STR ''unknownAssumption'') {
      u \<leftarrow> xml2unknown_input;
      xml_return (Assume_Unknown u [])
    } XMLor XMLdo (STR ''unknownProof'') {
      desc \<leftarrow> xml_text (STR ''description'');
      b \<leftarrow> xml2unknown_input;
      c \<leftarrow>* xml2subproof;
      xml_return (Assume_Unknown b c)
    };
    xml_return ret
  } x"
| "xml2subproof x =
    XMLdo (STR ''subProof'') {
      XMLdo {
        inp \<leftarrow> xml2dp_input' xml2name True;
        prf \<leftarrow> xml2dp_termination_proof;
        xml_return (Finite_assm_proof inp prf)
      } XMLor XMLdo {
        inp \<leftarrow> xml2inn_fp_trs_assm xml2name;
        case inp
        of Inl qtrs \<Rightarrow>
          XMLdo {
            prf \<leftarrow> xml2trs_termination_proof;
            xml_return (SN_assm_proof qtrs prf)
          }
        | Inr fptrs \<Rightarrow>
          XMLdo {
            prf \<leftarrow> xml2fptrs_termination_proof;
            xml_return (SN_FP_assm_proof fptrs prf)
          }
      }
    } x"

lemma drop_size_xml[dest]:
  assumes y: "drop n z = y" shows "sum_list (map size_xml y) \<le> sum_list (map size_xml z)"
proof (fold y, induct z arbitrary:n y)
  case Nil
  then show ?case by simp
next
  case (Cons a z)
  from le_trans[OF this le_add2] show ?case by (cases n, auto)
qed


partial_function (sum_bot)
  xml2complexity_proof :: "('f, 'l, string) complexity_proof xmlt2"
where
  [code]: "xml2complexity_proof x = xml_singleton (STR ''complexityProof'') (
    XMLdo (STR ''ruleShifting'') {
      rp \<leftarrow> CPF_Parser.ordering_constraint_proof xml2name True;
      del \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      ur \<leftarrow>? xml_singleton (STR ''usableRules'') (XTC_Parser.rules xml2name) id;
      prf \<leftarrow> xml2complexity_proof;
      xml_return (Rule_Shift_Complexity rp del ur prf)
    } XMLor XMLdo (STR ''usableRules'') {
      a \<leftarrow> xml_singleton (STR ''nonUsableRules'') (XTC_Parser.rules xml2name) id;
      b \<leftarrow> xml2complexity_proof;
      xml_return (Usable_Rules_Complexity a b)
    } XMLor XMLdo (STR ''split'') {
      a \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      b \<leftarrow> xml2complexity_proof;
      c \<leftarrow> xml2complexity_proof;
      xml_return (Split_Complexity a b c)
    } XMLor XMLdo (STR ''removeNonApplicableRules'') {
      a \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      b \<leftarrow> xml2complexity_proof;
      xml_return (Remove_Nonapplicable_Rules_Complexity a b)
    } XMLor xml_change (bounds_info xml2name) (xml_return \<circ> Matchbounds_Complexity)
    XMLor XMLdo (STR ''relativeBounds'') {
      a \<leftarrow> bounds_info xml2name;
      b \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      c \<leftarrow> xml2complexity_proof;
      xml_return (Matchbounds_Rel_Complexity a b c)
    } XMLor XMLdo (STR ''rIsEmpty'') {
      xml_return RisEmpty_Complexity
    } XMLor XMLdo (STR ''dtTransformation'') {
      S \<leftarrow> CPF_Parser.rule_pairs xml2name (STR ''strictDTs'') (STR ''ruleWithDT'');
      W \<leftarrow> CPF_Parser.rule_pairs xml2name (STR ''weakDTs'') (STR ''ruleWithDT'');
      inn \<leftarrow> innermostLhss xml2name;
      p \<leftarrow> xml2complexity_proof;
      xml_return (DT_Transformation (DT_Transformation_Info S W inn) p)
    } XMLor XMLdo (STR ''wdpTransformation'') {
      Comp \<leftarrow> XTC_Parser.symbols xml2name (STR ''compoundSymbols'');
      S \<leftarrow> CPF_Parser.rule_pairs xml2name (STR ''strictWDPs'') (STR ''ruleWithWDP'');
      W \<leftarrow> CPF_Parser.rule_pairs xml2name (STR ''weakWDPs'') (STR ''ruleWithWDP'');
      Q' \<leftarrow> innermostLhss xml2name;
      p \<leftarrow> xml2complexity_proof;
      xml_return (WDP_Transformation (WDP_Trans_Info (set Comp) S W Q') p)
    } XMLor XMLdo (STR ''unknownProof'') {
      a \<leftarrow> xml_text (STR ''description'');
      b \<leftarrow> xml2complexity_input' xml2name;
      c \<leftarrow>* XMLdo (STR ''subProof'') {
        inp \<leftarrow> xml2complexity_input' xml2name;
        prf \<leftarrow> xml2complexity_proof;
        xml_return (Complexity_assm_proof inp prf)
      };
      xml_return (Complexity_Assumption b c)
    } XMLor XMLdo (STR ''complexityAssumption'') {
       p \<leftarrow> xml2complexity_input' xml2name;
       xml_return (Complexity_Assumption p [])
    }
  ) id x"


partial_function (sum_bot) xml2ac_dp_termination_proof :: "(_, string) ac_dp_termination_proof xmlt2"
where [code]: "xml2ac_dp_termination_proof x = xml_singleton (STR ''acDPTerminationProof'') (
    XMLdo (STR ''acRedPairProc'') {
      a \<leftarrow> CPF_Parser.ordering_constraint_proof xml2name False;
      b \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
      c \<leftarrow> xml_singleton (STR ''usableRules'') (XTC_Parser.rules xml2name) id;
      d \<leftarrow> xml2ac_dp_termination_proof;
      xml_return (AC_Redpair_UR_Proc a b c d)
    } XMLor XMLdo (STR ''acSubtermProc'') {
      pi \<leftarrow> CPF_Parser.multiset_af xml2name;
      r \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
      prf \<leftarrow> xml2ac_dp_termination_proof;
      xml_return (AC_Subterm_Proc pi r prf)
    } XMLor XMLdo (STR ''acTrivialProc'') {
      xml_return AC_P_is_Empty
    } XMLor XMLdo (STR ''acMonoRedPairProc'') {
      a \<leftarrow> CPF_Parser.ordering_constraint_proof xml2name False;
      b \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
      c \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      d \<leftarrow> xml_singleton (STR ''usableRules'') (XTC_Parser.rules xml2name) id;
      e \<leftarrow> xml2ac_dp_termination_proof;
      xml_return (AC_Mono_Redpair_UR_Proc a b c d e)
    } XMLor XMLdo (STR ''acDepGraphProc'') {
      ret \<leftarrow>* XMLdo (STR ''component'') {
        dps \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
        _ \<leftarrow>? xml_bool (STR ''realScc''); \<comment> \<open>redundant\<close>
        prfOpt \<leftarrow>? xml2ac_dp_termination_proof;
        xml_return (prfOpt, dps)
      };
      xml_return (AC_Dep_Graph_Proc ret)
    }
   ) id x"

partial_function (sum_bot) xml2ac_termination_proof :: "('f, 'l, string) ac_termination_proof xmlt2"
where [code]: "xml2ac_termination_proof x = xml_singleton (STR ''acTerminationProof'') (
    XMLdo (STR ''acDependencyPairs'') {
      e \<leftarrow> xml_singleton (STR ''equations'') (XTC_Parser.rules xml2name) id;
      dpe \<leftarrow> xml_singleton (STR ''dpEquations'') (XTC_Parser.rules xml2name) id;
      dp \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
      rext \<leftarrow> xml_singleton (STR ''extensions'') (XTC_Parser.rules xml2name) id;
      p1 \<leftarrow> xml2ac_dp_termination_proof;
      p2opt \<leftarrow>? xml2ac_dp_termination_proof;
      case p2opt
      of Some p2 \<Rightarrow> xml_return (AC_DP_Trans (AC_dependency_pairs_proof e dp dpe rext) p1 p2)
      | None \<Rightarrow> xml_return (AC_DP_Trans_Single (AC_dependency_pairs_proof e dp dpe rext) p1)
    } XMLor XMLdo (STR ''acRuleRemoval'') {
      a \<leftarrow> CPF_Parser.ordering_constraint_proof xml2name False;
      b \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      c \<leftarrow> xml2ac_termination_proof;
      xml_return (AC_Rule_Removal a b c)
    } XMLor XMLdo (STR ''acRIsEmpty'') {
      xml_return AC_R_is_Empty
    }) id x"

definition nonloop :: "(_,string) non_loop_prf xmlt2" where
  "nonloop = XMLdo (STR ''nonLoop'') {
    a \<leftarrow> CPF_Parser.pat_rule_prf xml2name;
    b \<leftarrow> CPF_Parser.subst xml2name;
    c \<leftarrow> CPF_Parser.subst xml2name;
    d \<leftarrow> xml_nat (STR ''natural'');
    e \<leftarrow> xml_nat (STR ''natural'');
    f \<leftarrow> XTC_Parser.pos;
    xml_return (Non_loop_prf a b c d e f)
  }"

definition string :: "_ list xmlt2" where
  "string = xml_many (STR ''string'') xml2name id"

definition oc_srs :: "_ srs_rule xmlt2" where
  "oc_srs = xml_pair (STR ''overlapClosureSRS'') string string Pair"

definition word_pattern :: "_ word_pat xmlt2" where
  "word_pattern =
    XMLdo (STR ''wordPattern'') {
      l \<leftarrow> string;
      m \<leftarrow> string;
      f \<leftarrow> xml_nat (STR ''factor'');
      c \<leftarrow> xml_nat (STR ''constant'');
      r \<leftarrow> string;
      xml_return (l,(f,c,m),r)
  }"

definition derivation_pattern :: "_ deriv_pat xmlt2" where
  "derivation_pattern = xml_pair (STR ''derivationPattern'')
     word_pattern word_pattern Pair"

definition derivation_pattern_proof :: "_ dp_proof_step xmlt2" where
  "derivation_pattern_proof =
    (let
      oc = oc_srs;
      dp = derivation_pattern;
      s = string in
    xml_singleton (STR ''derivationPatternProof'') (
      xml_pair (STR ''OC1'') oc (xml_bool (STR ''isPair'')) OC1
    XMLor xml_tuple6 (STR ''OC2'') oc oc oc s s s OC2
    XMLor xml_tuple6 (STR ''OC2prime'') oc oc oc s s s OC2p
    XMLor xml_tuple5 (STR ''OC3'') oc oc oc s s  OC3
    XMLor xml_tuple5 (STR ''OC3prime'') oc oc oc s s OC3p
    XMLor xml_pair (STR ''OCintoDP1'') dp oc OCDP1
    XMLor xml_pair (STR ''OCintoDP2'') dp oc OCDP2
    XMLor xml_pair (STR ''equivalent'') dp dp WPEQ
    XMLor xml_pair (STR ''lift'') dp dp Lift
    XMLor xml_tuple5 (STR ''DP_OC_1_1'') dp dp oc s s DPOC1_1
    XMLor xml_tuple6 (STR ''DP_OC_1_2'') dp dp oc s s s DPOC1_2
    XMLor xml_tuple5 (STR ''DP_OC_2'') dp dp oc s s DPOC2
    XMLor xml_tuple5 (STR ''DP_OC_3_1'') dp dp oc s s DPOC3_1
    XMLor xml_tuple6 (STR ''DP_OC_3_2'') dp dp oc s s s DPOC3_2
    XMLor xml_tuple5 (STR ''DP_DP_1_1'') dp dp dp s s DPDP1_1
    XMLor xml_tuple5 (STR ''DP_DP_1_2'') dp dp dp s s DPDP1_2
    XMLor xml_tuple5 (STR ''DP_DP_2_1'') dp dp dp s s DPDP2_1
    XMLor xml_tuple5 (STR ''DP_DP_2_2'') dp dp dp s s DPDP2_2
  ) id)"

definition nonloop_srs_reason :: "(_ dp_proof_step list \<Rightarrow> _ non_loop_srs_proof)xmlt2" where
  "nonloop_srs_reason =
    xml_triple (STR ''selfEmbeddingOC'')
        string string string
        (\<lambda> l m r. SE_OC (m, l @ m @ r) l r)
    XMLor xml_triple (STR ''selfEmbeddingDP'')
        derivation_pattern string string
        SE_DP"

definition nonloop_srs :: "_ non_loop_srs_proof xmlt2" where
  "nonloop_srs = xml_pair (STR ''nonterminatingSRS'')
    (xml_many (STR ''derivationPatterns'') derivation_pattern_proof id)
    nonloop_srs_reason
    (\<lambda> list prf. prf list)"

definition not_wn_ta :: "(_, string) not_wn_ta_prf xmlt2" where
  "not_wn_ta = xml_pair (STR ''notWNTreeAutomaton'')
    (tree_automaton (ta_normal_lhs xml2name))
    closed_criterion
    (\<lambda> ta rel. Not_wn_ta_prf ta rel)"

definition xml2inn_fp_nt_trs_assm ::
    "(('f, 'l, string) qtrsLL + ('f, 'l, string) fptrsLL) xmlt2"
where
  "xml2inn_fp_nt_trs_assm = xml_change (xml2_trs_input xml2name) (\<lambda> inp.
   case inp of
     Inn_TRS_input inn r s start \<Rightarrow>
      if s \<noteq> [] then
        xml_error (STR ''relative rules are not allowed here'')
      else if start \<noteq> Full then
        xml_error (STR ''start terms not allowed here'')
      else
        xml_return (Inl (default_nfs_nt_trs, strategy_to_Q inn r, r))
    | FP_TRS_input fp r \<Rightarrow> xml_return (Inr (strategy_to_fp fp r, r))
    | _ \<Rightarrow> xml_error (STR ''non-relative TRS expected'')
  )"

definition xml2fp_nt_trs_assm :: "('f, 'l, string) fptrsLL xmlt2"
where
  "xml2fp_nt_trs_assm = xml_change (xml2_trs_input xml2name) (\<lambda>inp.
    case inp of
      FP_TRS_input fp r \<Rightarrow> xml_return (strategy_to_fp fp r, r)
    | _ \<Rightarrow> xml_error (STR ''outermost/forbidden-pattern/context-sensitive TRS expected'')
  )"

definition xml2inn_nt_trs_assm where
  "xml2inn_nt_trs_assm = xml_change (xml2_trs_input xml2name) (\<lambda>inp.
    case inp of Inn_TRS_input inn r s start \<Rightarrow>
      if s \<noteq> [] then xml_error (STR ''(innermost) TRS without relative rules expected'')
      else if start \<noteq> Full then xml_error (STR ''start term is not allowed here'')
      else xml_return (default_nfs_nt_trs, strategy_to_Q inn r, r)
   )"

definition xml2inn_rel_nt_trs_assm where
  "xml2inn_rel_nt_trs_assm = xml_change (xml2_trs_input xml2name) (\<lambda>inp.
    case inp of Inn_TRS_input inn r s start \<Rightarrow>
    if start \<noteq> Full then xml_error (STR ''start term is not allowed here'')
    else xml_return (default_nfs_nt_trs, strategy_to_Q inn r, r, s)
   )"
(*
abbreviation sub_disproof where "sub_disproof xml2name
  xml2trs_nontermination_proof
  xml2dp_nontermination_proof
  xml2reltrs_nontermination_proof
  xml2fp_nontermination_proof
  xml2unknown_disproof
  x \<equiv>
    let cs = Xml.children x
     in (if length cs \<noteq> 2 then Xmlt.fail ''subProof'' x
      else let inp = take 2 (Xml.tag (cs ! 1)) in
      (if inp = ''tr'' then
        do {
          io_trs <- xml2inn_fp_nt_trs_assm xml2name (hd cs);
          (case io_trs of
            Inl qtrs \<Rightarrow> Xmlt.change xml2trs_nontermination_proof (\<lambda> prf. Not_SN_assm_proof qtrs prf :: ('f, 'l, string, 'a, 'b, 'c, 'd,'e) generic_assm_proof)
          | Inr fptrs \<Rightarrow> Xmlt.change xml2fp_nontermination_proof (\<lambda> prf. Not_SN_FP_assm_proof fptrs prf)) (cs ! 1)
        }
      else if inp = ''dp'' then
        Xmlt.pair ''subProof'' (xml2dp_input' False xml2name) xml2dp_nontermination_proof (\<lambda> qdp prf. Infinite_assm_proof qdp prf :: ('f, 'l, string, 'a, 'b, 'c, 'd,'e) generic_assm_proof) x
      else if inp = ''re'' then
        Xmlt.pair ''subProof'' (xml2inn_rel_nt_trs_assm xml2name) xml2reltrs_nontermination_proof Not_RelSN_assm_proof x
      else if inp = ''un'' then
        Xmlt.pair ''subProof'' xml2unknown_input xml2unknown_disproof Unknown_assm_proof x
      else Xmlt.fail ''subProof'' x))"
*)

partial_function_mr (sum_bot) xml2dp_nontermination_proof ::
    "('f, 'l, string) dp_nontermination_proof xmlt2" and
  xml2trs_nontermination_proof ::
    "('f, 'l, string) trs_nontermination_proof xmlt2" and
  xml2reltrs_nontermination_proof ::
    "('f, 'l, string) reltrs_nontermination_proof xmlt2" and
  xml2unknown_disproof :: "('f, 'l, string) neg_unknown_proof xmlt2"
  where
  "xml2dp_nontermination_proof x = xml_singleton (STR ''dpNonterminationProof'') (
    XMLdo (STR ''dpRuleRemoval'') {
      p \<leftarrow>? xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
      r \<leftarrow>? xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      prf \<leftarrow> xml2dp_nontermination_proof;
      xml_return (DP_Rule_Removal (Rule_removal_nonterm_dp_prf p r) prf)
    } XMLor XMLdo (STR ''loop'') {
      (s,rseq) \<leftarrow> CPF_Parser.relsteps xml2name;
      \<sigma> \<leftarrow> CPF_Parser.subst xml2name;
      c \<leftarrow> CPF_Parser.ctxt xml2name;
      xml_return (DP_Loop (DP_loop_prf s rseq \<sigma> c))
    } XMLor xml_change (CPF_Parser.nonloop xml2name)
      (xml_return \<circ> DP_Nonloop)
    XMLor XMLdo (STR ''innermostLhssRemovalProc'') {
      q \<leftarrow> innermostLhss xml2name;
      p \<leftarrow> xml2dp_nontermination_proof;
      xml_return (DP_Q_Reduction (DP_q_reduction_nonterm_prf q) p)
    } XMLor XMLdo (STR ''innermostLhssIncreaseProc'') {
      q \<leftarrow> innermostLhss xml2name;
      p \<leftarrow> xml2dp_nontermination_proof;
      xml_return (DP_Q_Increase (Q_increase_nonterm_dp_prf q) p)
    } XMLor XMLdo (STR ''instantiationProc'') {
      p \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
      prf \<leftarrow> xml2dp_nontermination_proof;
      xml_return (DP_Instantiation (Instantiation_complete_proc_prf p) prf)
    } XMLor XMLdo (STR ''narrowingProc'') {
      st \<leftarrow> XTC_Parser.rule xml2name;
      po \<leftarrow> XTC_Parser.pos;
      p \<leftarrow> xml_singleton (STR ''narrowings'') (XTC_Parser.rules xml2name) id;
      prf \<leftarrow> xml2dp_nontermination_proof;
      xml_return (DP_Narrowing (Narrowing_complete_proc_prf st po p) prf)
    } XMLor XMLdo (STR ''rewritingProc'') {
      (s,t) \<leftarrow> XTC_Parser.rule xml2name;
      (p,lr,t') \<leftarrow> CPF_Parser.rstep xml2name;
      c \<leftarrow>? XTC_Parser.rule xml2name;
      Uopt \<leftarrow>? xml_singleton (STR ''usableRules'') (XTC_Parser.rules xml2name) id;
      prf \<leftarrow> xml2dp_nontermination_proof;
      let st'' = (case c of Some v \<Rightarrow> v | None \<Rightarrow> (s,t'));
      xml_return (DP_Rewriting (Rewriting_complete_proc_prf Uopt (s,t) (s,t') st'' lr p) prf)
    } XMLor XMLdo (STR ''switchFullStrategyProc'') {
      a \<leftarrow> CPF_Parser.wcr_proof xml2name;
      b \<leftarrow> xml2dp_nontermination_proof;
      xml_return (DP_Termination_Switch a b)
    } XMLor XMLdo (STR ''infinitenessAssumption'') {
      qdp \<leftarrow> xml2dp_input' xml2name False;
      xml_return (DP_Assume_Infinite qdp [])
    } XMLor XMLdo (STR ''unknownProof'') {
      a \<leftarrow> xml_text (STR ''description'');
      b \<leftarrow> xml2dp_input' xml2name False;
      c \<leftarrow>* xml_any;
      xml_return (DP_Assume_Infinite b [])
  }) id x"
| "xml2trs_nontermination_proof x =
  xml_singleton (STR ''trsNonterminationProof'') (
  XMLdo (STR ''variableConditionViolated'') {
    xml_return TRS_Not_Well_Formed
  } XMLor xml_change (CPF_Parser.loop xml2name)
    (\<lambda>(s, rseq, \<sigma>, C). xml_return (TRS_Loop (TRS_loop_prf s (map (\<lambda>(x, y, _, z). (x, y, z)) rseq) \<sigma> C)))
  XMLor xml_change (CPF_Parser.nonloop xml2name) (xml_return \<circ> TRS_Nonloop)
  XMLor xml_change (CPF_Parser.nonloop_srs xml2name) (xml_return \<circ> TRS_Nonloop_SRS)
  XMLor XMLdo (STR ''rightGroundNontermination'') {
      xml_return TRS_Not_RG_Decision 
  } XMLor XMLdo (STR ''ruleRemoval'') {
    r \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
    p \<leftarrow> xml2trs_nontermination_proof;
    xml_return (TRS_Rule_Removal (Rule_removal_nonterm_trs_prf r) p)
  } XMLor XMLdo (STR ''dpTrans'') {
    p \<leftarrow> xml_singleton (STR ''dps'') (XTC_Parser.rules xml2name) id;
    _ \<leftarrow> xml_bool (STR ''markedSymbols'');
    c \<leftarrow> xml2dp_nontermination_proof;
    xml_return (TRS_DP_Trans (DP_trans_nontermination_tt_prf p) c)
  } XMLor XMLdo (STR ''stringReversal'') {
    _ \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
    p \<leftarrow> xml2trs_nontermination_proof;
    xml_return (TRS_String_Reversal p)
  } XMLor XMLdo (STR ''constantToUnary'') {
     v \<leftarrow> XTC_Parser.plain_var;
     ren \<leftarrow> CPF_Parser.renaming xml2name;
     S \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
     p \<leftarrow> xml2trs_nontermination_proof;
     xml_return (TRS_Constant_String (Const_string_complete_proof v ren S) p)
  } XMLor XMLdo (STR ''innermostLhssIncrease'') {
    q \<leftarrow> innermostLhss xml2name;
    r \<leftarrow> xml2trs_nontermination_proof;
    xml_return (TRS_Q_Increase (Q_increase_nonterm_trs_prf q) r)
  } XMLor XMLdo (STR ''switchFullStrategy'') {
    a \<leftarrow> CPF_Parser.wcr_proof xml2name;
    b \<leftarrow> xml2trs_nontermination_proof;
    xml_return (TRS_Termination_Switch a b)
  } XMLor XMLdo (STR ''uncurry'') {
      i \<leftarrow> CPF_Parser.uncurry_info xml2name;
      r \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      p \<leftarrow> xml2trs_nontermination_proof;
      xml_return (TRS_Uncurry (Uncurry_nt_proof i r) p)
  } XMLor xml_change (CPF_Parser.not_wn_ta xml2name)
    (xml_return \<circ> TRS_Not_WN_Tree_Automaton)
  XMLor XMLdo (STR ''nonterminationAssumption'') {
    qtrs \<leftarrow> xml2inn_nt_trs_assm;
    xml_return (TRS_Assume_Not_SN qtrs [])
  } XMLor XMLdo (STR ''unknownProof'') {
    a \<leftarrow> xml_text (STR ''description'');
    b \<leftarrow> xml2inn_nt_trs_assm;
    c \<leftarrow>* xml_any;
    xml_return (TRS_Assume_Not_SN b [])
  }) id x"
| "xml2reltrs_nontermination_proof x =
  xml_singleton (STR ''relativeNonterminationProof'') (
  XMLdo (STR ''variableConditionViolated'') {
    xml_return Rel_Not_Well_Formed
  } XMLor xml_change (CPF_Parser.loop xml2name)
      (\<lambda>(s, rseq, \<sigma>, C). xml_return (Rel_Loop (Rel_trs_loop_prf s rseq \<sigma> C)))
  XMLor xml_change xml2trs_nontermination_proof (xml_return \<circ> Rel_R_Not_SN)
  XMLor XMLdo (STR ''ruleRemoval'') {
    r \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) Some;
    s \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) Some;
    p \<leftarrow> xml2reltrs_nontermination_proof;
    xml_return (Rel_Rule_Removal (Rule_removal_nonterm_reltrs_prf r s) p)
  } XMLor XMLdo (STR ''stringReversal'') {
    _ \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) Some;
    _ \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) Some;
    p \<leftarrow> xml2reltrs_nontermination_proof;
    xml_return (Rel_TRS_String_Reversal p)
  } XMLor XMLdo (STR ''nonterminationAssumption'') {
    qtrs \<leftarrow> xml2inn_rel_nt_trs_assm;
    xml_return (Rel_TRS_Assume_Not_SN qtrs [])
  } XMLor XMLdo (STR ''unknownProof'') {
      a \<leftarrow> xml_text (STR ''description'');
      b \<leftarrow> xml2inn_rel_nt_trs_assm;
      c \<leftarrow>* xml_any;
      xml_return (Rel_TRS_Assume_Not_SN b [])
  }) id x"
| "xml2unknown_disproof x = xml_singleton (STR ''unknownInputProof'') (
    xml_singleton (STR ''unknownAssumption'')
        xml2unknown_input
        (\<lambda> u. Assume_NT_Unknown u [])
    XMLor XMLdo (STR ''unknownProof'') {
        a \<leftarrow> xml_text (STR ''description'');
        b \<leftarrow> xml2unknown_input;
        c \<leftarrow>* xml_any;
        xml_return (Assume_NT_Unknown b [])
  }) id x"


definition rule_labeling_function :: "(_, string) rule_lab_repr xmlt2"
where
  "rule_labeling_function = xml_many (STR ''ruleLabelingFunction'')
   (xml_pair (STR ''ruleLabelingFunctionEntry'')
    (XTC_Parser.rule xml2name)
    (xml_nat (STR ''label''))
    Pair) id"

fun add_source_lab_proof
where
  "add_source_lab_proof (Rule_Labeling rl js _) _ prf = Rule_Labeling rl js (Some prf)"
| "add_source_lab_proof (Rule_Labeling_Conv rl cs _) (Some n) prf = Rule_Labeling_Conv rl cs (Some (n, prf))"

fun nat_or_empty (*TODO: ad-hoc, syntax should be changed*)
where "nat_or_empty ([],s) = Right None"
| "nat_or_empty ([XML_text txt],s) = (
  case int_of_string txt of Inr n \<Rightarrow>
    Right (Some (Int.nat n))
  | Inl err \<Rightarrow> xml_error err ([]::xml list,s)
  )"

definition xml2ms_signature
  where "xml2ms_signature = xml_many (STR ''manySortedSignature'') (
       xml_triple (STR ''manySortedFunction'')
         xml2name
         (xml_many (STR ''args'') (xml_text (STR ''sort'')) id)
         (xml_singleton (STR ''result'') (xml_text (STR ''sort'')) id)
       (\<lambda>a b c. (a, b, c))
    ) id"

partial_function (sum_bot) xml2cr_proof :: "(_,_,string) cr_proof xmlt2"
where [code]: "xml2cr_proof x = xml_singleton (STR ''crProof'') (
    xml_pair (STR ''wcrAndSN'')
        (wcr_proof xml2name)
        xml2trs_termination_proof
      SN_WCR
    XMLor xml_leaf (STR ''orthogonal'') Weakly_Orthogonal
    XMLor xml_change (xml_nat (STR ''stronglyClosed'')) (xml_return \<circ> Strongly_Closed)
    XMLor xml_change (xml_do (STR ''parallelClosed'') nat_or_empty) (xml_return \<circ> Parallel_Closed)
    XMLor xml_change (xml_do (STR ''developmentClosed'') nat_or_empty) (xml_return \<circ> Development_Closed)
    XMLor xml_triple (STR ''criticalPairClosingSystem'')
        (xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id)
        xml2trs_termination_proof
        (xml_nat (STR ''nrSteps''))
      Critical_Pair_Closing_System
    XMLor xml_pair (STR ''ruleLabeling'')
        (CPF_Parser.rule_labeling_function xml2name)
        (CPF_Parser.joinable_critical_pairs xml2name)
      (\<lambda> rl js. Rule_Labeling rl js None)
    XMLor XMLdo (STR ''decreasingDiagrams'') {
        a \<leftarrow>? xml2trs_termination_proof;
        ret \<leftarrow>
          XMLdo (STR ''ruleLabeling'') {
            rl \<leftarrow> CPF_Parser.rule_labeling_function xml2name;
            js \<leftarrow> CPF_Parser.joinable_critical_pairs xml2name;
            case a of
              Some prf \<Rightarrow> xml_return (add_source_lab_proof (Rule_Labeling rl js None) None prf)
            | _ \<Rightarrow> xml_return (Rule_Labeling rl js None)
          } XMLor XMLdo (STR ''ruleLabelingConv'') {
            rl \<leftarrow> CPF_Parser.rule_labeling_function xml2name;
            cs \<leftarrow> CPF_Parser.convertible_critical_peaks xml2name;
            case a of
              Some prf \<Rightarrow> XMLdo {
                n \<leftarrow> xml_nat (STR ''nrSteps'');
                xml_return (add_source_lab_proof (Rule_Labeling_Conv rl cs None) (Some n) prf)
              }
            | _ \<Rightarrow> xml_return (Rule_Labeling_Conv rl cs None)
          };
        xml_return ret
    } XMLor XMLdo (STR ''redundantRules'') {
      trs \<leftarrow> xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id;
      n \<leftarrow> xml_nat (STR ''nrSteps'');
      cs \<leftarrow>[[]] xml_many (STR ''conversions'') (CPF_Parser.conversion xml2name) (map snd);
      prf \<leftarrow> xml2cr_proof;
      xml_return (Redundant_Rules trs n cs prf)
    } XMLor XMLdo (STR ''persistentDecomposition'') {
      sig \<leftarrow> xml2ms_signature;
      subs \<leftarrow>* xml_pair (STR ''component'')
       (xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id)
       xml2cr_proof
       Pair;
      xml_return (Persistent_Decomposition sig subs)
  }) id x"

definition compatible_ta :: "ltag \<Rightarrow> ((string,_)tree_automaton \<times> string ta_relation) xmlt2" where
  "compatible_ta tag = XMLdo tag {
    a \<leftarrow> tree_automaton (ta_normal_lhs xml2name);
    b \<leftarrow>[Id_Relation] closed_criterion;
    xml_return (a,b)
  }"

definition xml2completion_proof :: "(_,_,string) completion_proof xmlt2"
where "xml2completion_proof = XMLdo (STR ''completionProof'') {
     w \<leftarrow> wcr_proof xml2name;
     t \<leftarrow> xml2trs_termination_proof;
     (s1,s2) \<leftarrow> XMLdo (STR ''equivalenceProof'') {
       s1 \<leftarrow> CPF_Parser.subsumption_proof xml2name;
       s2 \<leftarrow>? CPF_Parser.subsumption_proof xml2name;
       xml_return (s1,s2)
      };
     xml_return (SN_WCR_Eq w t s1 s2)
  }"

definition xml2ordered_completion_step :: "(_,string) oc_irule xmlt2"
  where "xml2ordered_completion_step =
      xml_triple (STR ''deduce'')
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (\<lambda> s t u. OC_Deduce s t u)
    XMLor
      (xml_pair (STR ''orientl'')
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (\<lambda>s t. OC_Orientl s t))
    XMLor
      (xml_pair (STR ''orientr'')
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (\<lambda>s t. OC_Orientr s t))
    XMLor
      (xml_singleton (STR ''delete'')
        (XTC_Parser.term xml2name)
        (\<lambda>s. OC_Delete s))
    XMLor
      (xml_triple (STR ''compose'')
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (\<lambda>s t u. OC_Compose s t u))
    XMLor
      (xml_triple (STR ''simplifyl'')
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (\<lambda>s t u. OC_Simplifyl s t u))
    XMLor
      (xml_triple (STR ''simplifyr'')
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (\<lambda>s t u. OC_Simplifyr s t u))
    XMLor
      (xml_triple (STR ''collapse'')
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (\<lambda>s t u. OC_Collapse s t u))
"

definition "xml2ordered_completion_proof =
      xml_singleton (STR ''orderedCompletionProof'') (
        xml_many (STR ''run'')
          (xml_singleton (STR ''orderedCompletionStep'') xml2ordered_completion_step id)
          id
      ) OKB"

definition "unraveling_info =
  (xml_many (STR ''unravelingInformation'') (XMLdo (STR ''unravelingEntry'') {
     a \<leftarrow> XTC_Parser.crule xml2name (STR ''conditionalRule'');
     b \<leftarrow>* XTC_Parser.rule xml2name;
     xml_return (a,b)
  }) id)"

definition "xml2quasi_reductive_proof = xml_singleton (STR ''quasiReductiveProof'') (
  xml_pair (STR ''unraveling'')
    (CPF_Parser.unraveling_info xml2name)
    xml2trs_termination_proof
    Unravel
  ) id"

definition "xml2state_map = xml_many (STR ''stateMap'')
  (xml_pair (STR ''entry'') state (XTC_Parser.term xml2name) Pair)
  (\<lambda>xs x. the (map_of xs x))"

definition "xml2nonreachable_etac_info = xml_tuple5 (STR ''nonreachableEtac'')
  (CPFsignature xml2name)
  xml2name
  xml2name
  (tree_automaton (ta_normal_lhs xml2name))
  (xml2state_map)
  (\<lambda>F a c A m. Nonreachable_ETAC F a c (map_states_impl m A))"

definition "xml2ordered_completion = XMLdo (STR ''orderedCompletion'') {
  (rs, es, ro) \<leftarrow> xml2ordered_completion_result xml2name;
  p \<leftarrow> xml2ordered_completion_proof;
  xml_return (rs, es, ro, p)
}"

partial_function (sum_bot) xml2nonreachability_proof :: "(_, _) nonreachability_proof xmlt2"
where [code]: "xml2nonreachability_proof x = xml_singleton (STR ''nonreachabilityProof'') (
    xml_leaf (STR ''nonreachableTcap'') Nonreachable_Gtcap
    XMLor xml2nonreachable_etac_info
    XMLor xml_pair (STR ''nonreachableSubstApprox'')
      (XTC_Parser.rules xml2name)
      xml2nonreachability_proof
      Nonreachable_Subst_Approx
    XMLor xml_singleton (STR ''nonreachableReverse'') xml2nonreachability_proof
      Nonreachable_Reverse
    XMLor (XMLdo (STR ''nonreachableFGCR'') {
      eq \<leftarrow> XMLdo (STR ''eqSymbol'') {x \<leftarrow> xml2name; xml_return x};
      tr \<leftarrow> XMLdo (STR ''trueSymbol'') {x \<leftarrow> xml2name; xml_return x};
      fa \<leftarrow> XMLdo (STR ''falseSymbol'') {x \<leftarrow> xml2name; xml_return x};
      (rs, es, ro, p) \<leftarrow> xml2ordered_completion;
      xml_return (Nonreachable_FGCR eq tr fa es rs ro p)
    })
  ) id x"


definition "xml2nonjoinability_proof = xml_singleton (STR ''nonjoinabilityProof'') (
  xml_leaf (STR ''nonjoinableTcap'') Nonjoinable_Tcap
  XMLor xml_change xml2nonreachability_proof (xml_return \<circ> Nonjoinable_Ground_NF)
  ) id"

definition xml2inline_cond_info
  where "xml2inline_cond_info = XMLdo (STR ''inlinedRules'') {
    ret \<leftarrow>* XMLdo (STR ''inlinedRule'') {
      rule \<leftarrow> XTC_Parser.crule xml2name (STR ''rule'');
      conds \<leftarrow> XTC_Parser.conditions xml2name (STR ''inlinedConditions'');
      xml_return (rule,conds)
    };
    xml_return ret
  }"

partial_function (sum_bot) xml2infeasibility_proof :: "(_,_) infeasibility_proof xmlt2"
where [code]: "xml2infeasibility_proof x = xml_singleton (STR ''infeasibilityProof'') (
    xml_pair (STR ''infeasibleCompoundConditions'')
      xml2name
      xml2nonreachability_proof
      Infeasible_Compound_Conditions
    XMLor xml_pair (STR ''infeasibleEquation'')
      (XTC_Parser.rule xml2name)
      xml2nonreachability_proof
      (\<lambda>(l, r). Infeasible_Equation l r)
    XMLor xml_pair (STR ''infeasibleSubset'')
      (XTC_Parser.rules xml2name)
      xml2infeasibility_proof
      Infeasible_Subset
    XMLor xml_tuple4 (STR ''infeasibleRhssEqual'')
      (XTC_Parser.term xml2name)
      (XTC_Parser.term xml2name)
      (XTC_Parser.term xml2name)
      xml2nonjoinability_proof
      Infeasible_Rhss_Equal
    XMLor xml_tuple4 (STR ''infeasibleTrans'')
      (XTC_Parser.term xml2name)
      (XTC_Parser.term xml2name)
      (XTC_Parser.term xml2name)
      xml2nonreachability_proof
      Infeasible_Trans
    XMLor xml_pair (STR ''ifritRules'')
      (XMLdo (STR ''rules'') { ret \<leftarrow>* XTC_Parser.crule xml2name (STR ''rule''); xml_return ret })
      xml2infeasibility_proof
      (Infeasible_Transform \<circ> Ifrit_Rules_Inf)
    XMLor xml_triple (STR ''leftInlineConditions'')
      (XMLdo (STR ''rules'') { ret \<leftarrow>* XTC_Parser.crule xml2name (STR ''rule''); xml_return ret })
      xml2inline_cond_info
      xml2infeasibility_proof
      (Infeasible_Transform \<circ>\<circ> Left_Inline_Conditions_Inf)
    XMLor xml_triple (STR ''rightInlineConditions'')
      (XMLdo (STR ''rules'') { ret \<leftarrow>* XTC_Parser.crule xml2name (STR ''rule''); xml_return ret })
      xml2inline_cond_info
      xml2infeasibility_proof
      (Infeasible_Transform \<circ>\<circ> Right_Inline_Conditions_Inf)
  ) id x"


definition "xml2ao_infeasibility_proof = xml_singleton (STR ''aoInfeasibilityProof'') (
  xml_change xml2infeasibility_proof (xml_return \<circ> AO_Infeasibility_Proof)
  XMLor xml_tuple4 (STR ''aoLhssEqual'')
    (XTC_Parser.term xml2name)
    (XTC_Parser.term xml2name)
    (XTC_Parser.term xml2name)
    xml2nonjoinability_proof
    AO_Lhss_Equal
  ) id"

definition xml2infeasible_conds
where
  "xml2infeasible_conds = xml_many (STR ''infeasibleConditions'')
    (xml_pair (STR ''infeasibleCondition'')
      (XTC_Parser.conditions xml2name (STR ''conditions''))
      xml2infeasibility_proof
      Pair)
    id"

definition xml2ao_infeasible_conds
where
  "xml2ao_infeasible_conds = xml_many (STR ''aoInfeasibleConditions'')
    (xml_triple (STR ''aoInfeasibleCondition'')
      (XTC_Parser.conditions xml2name (STR ''conditions''))
      (XTC_Parser.conditions xml2name (STR ''conditions''))
      xml2ao_infeasibility_proof
      (\<lambda>cs\<^sub>1 cs\<^sub>2 p. (cs\<^sub>1, cs\<^sub>2, p)))
    id"

partial_function_mr (sum_bot) cstep :: "(_, _) cstep_proof xmlt2"
     and csteps :: "(_, _) cstep_proof list xmlt2"
where
  "cstep x = xml_tuple6 (STR ''conditionalRewriteStep'')
    (XTC_Parser.crule xml2name (STR ''rule''))
    XTC_Parser.pos
    (CPF_Parser.subst xml2name)
    (xml_singleton (STR ''source'') (XTC_Parser.term xml2name) id)
    (xml_singleton (STR ''target'') (XTC_Parser.term xml2name) id)
    (xml_many (STR ''conditions'') csteps id)
    (\<lambda>r p s t u qs. Cstep_step r p (mk_subst Var s) t u qs) x"
| "csteps x = xml_many (STR ''conditionalRewritingSequence'') cstep id x"

definition xml2const_map :: "(_ \<Rightarrow> string option) xmlt2"
where
  "xml2const_map = xml_many (STR ''constMap'') (xml_pair (STR ''entry'')
    (xml_singleton (STR ''symbol'') xml2name id)
    (xml_text (STR ''const''))
    Pair)
    (\<lambda>xs. map_of xs)"

definition xml2context_joinable_proof :: "(_, _) context_joinable_proof xmlt2"
where
  "xml2context_joinable_proof = xml_tuple4 (STR ''contextJoinabilityProof'')
    (XTC_Parser.term xml2name)
    (CPF_Parser.csteps xml2name)
    (CPF_Parser.csteps xml2name)
    xml2const_map
    (\<lambda>t ps qs m. Contextual_Join (orig_term m t) (map (orig_cstep m) ps) (map (orig_cstep m) qs))"

definition xml2context_joinable_ccps ::
  "((_, string) term \<times> (_, string) term \<times> (_, string) rules \<times> (_, string) context_joinable_proof) list xmlt2"
where
  "xml2context_joinable_ccps = XMLdo (STR ''contextJoinableCCPs'') {
    ret \<leftarrow>* XMLdo (STR ''contextJoinableCCP'') {
      s \<leftarrow> XTC_Parser.term xml2name;
      t \<leftarrow> XTC_Parser.term xml2name;
      cs \<leftarrow> XTC_Parser.conditions xml2name (STR ''conditions'');
      p \<leftarrow> xml2context_joinable_proof;
      xml_return (s, t, cs, p)
    };
    xml_return ret
  }"

definition xml2unfeasible_proof :: "(_, string) unfeasible_proof xmlt2"
where
  "xml2unfeasible_proof = xml_tuple5 (STR ''unfeasibilityProof'')
    (xml_triple (STR ''terms'')
      (XTC_Parser.term xml2name)
      (XTC_Parser.term xml2name)
      (XTC_Parser.term xml2name)
    (\<lambda>t u v. (t, u, v)))
    csteps
    csteps
    (xml_pair (STR ''rules'') (XTC_Parser.crule xml2name (STR ''rule'')) (XTC_Parser.crule xml2name (STR ''rule'')) Pair)
    xml2const_map
    (\<lambda>(t, u, v) ps qs (r, r') m.
      UnfeasibleOverlap (orig_term m t) (orig_term m u) (orig_term m v)
        (map (orig_cstep m) ps) (map (orig_cstep m) qs) r r')"

definition xml2unfeasible_ccps ::
  "((_, string) subst \<times> (_, string) unfeasible_proof) list xmlt2"
where
  "xml2unfeasible_ccps = xml_many (STR ''unfeasibleCCPs'')
    (xml_pair (STR ''unfeasibleCCP'')
      (CPF_Parser.subst xml2name)
      xml2unfeasible_proof
      (\<lambda>s p. (subst_of s, p)))
    id"

definition xml2infeasible_rules_info
  where
    "xml2infeasible_rules_info = xml_many (STR ''infeasibleRules'')
      (xml_pair (STR ''infeasibleRule'')
        (XTC_Parser.crule xml2name (STR ''rule''))
        xml2infeasibility_proof
        Pair)
      id"

partial_function (sum_bot) xml2conditional_cr_proof ::
    "(_, _, string) conditional_cr_proof xmlt2"
where [code]: "xml2conditional_cr_proof x = xml_singleton (STR ''conditionalCrProof'') (
    xml_singleton (STR ''unconditional'')
      xml2cr_proof
      Unconditional_CR
    XMLor xml_pair (STR ''unraveling'')
      (CPF_Parser.unraveling_info xml2name)
      xml2cr_proof
      Unravel_CR
    XMLor xml_triple (STR ''inlineConditions'')
      (XMLdo (STR ''rules'') { ret \<leftarrow>* XTC_Parser.crule xml2name (STR ''rule''); xml_return ret })
      xml2inline_cond_info
      xml2conditional_cr_proof
      (Transformation_CR \<circ>\<circ> Inline_Conditions_CCRT)
    XMLor xml_pair (STR ''infeasibleRuleRemoval'')
      xml2infeasible_rules_info
      xml2conditional_cr_proof
      (Transformation_CR \<circ> Infeasible_Rule_Removal_CCRT)
    XMLor xml_leaf (STR ''almostOrthogonal'') Almost_Orthogonal_CR
    XMLor xml_singleton (STR ''almostOrthogonalModuloInfeasibility'')
        xml2ao_infeasible_conds Almost_Orthogonal_Modulo_Infeasibility_CR'
    XMLor xml_tuple4 (STR ''al94'')
      xml2quasi_reductive_proof
      xml2context_joinable_ccps
      xml2infeasible_conds
      xml2unfeasible_ccps
      AL94_CR
  ) id x"

end

hide_const (open) string
hide_const (open) nonloop_srs
hide_const (open) rule_labeling_function

partial_function (sum_bot) xml2non_join_info :: "string lt xmlt2 \<Rightarrow> (string lt,string,string) non_join_info xmlt2" where
  [code]: "xml2non_join_info xml2name x = (
    xml_leaf (STR ''distinctNormalForms'') Diff_NFs
    XMLor xml_leaf (STR ''capNotUnif'') Tcap_Not_Unif
    XMLor xml_pair (STR ''subterm'')
        XTC_Parser.pos
        (xml2non_join_info xml2name)
        Subterm_NJ
    XMLor xml_pair (STR ''grounding'')
        (CPF_Parser.subst xml2name)
        (xml2non_join_info xml2name)
        Grounding
    XMLor xml_pair (STR ''emptyTreeAutomataIntersection'')
        (compatible_ta xml2name (STR ''firstAutomaton''))
        (compatible_ta xml2name (STR ''secondAutomaton''))
        (\<lambda> (ta1,rel1) (ta2,rel2). Tree_Aut_Intersect_Empty ta1 rel1 ta2 rel2)
    XMLor xml_singleton (STR ''differentInterpretation'')
        (CPF_Parser.sl_variant xml2name)
        Finite_Model_Gt
    XMLor xml_singleton (STR ''strictDecrease'')
        (CPF_Parser.ordering_constraint_proof xml2name False)
        Reduction_Pair_Gt
    XMLor xml_pair (STR ''argumentFilterNonJoin'')
        (CPF_Parser.afs xml2name)
        (xml2non_join_info xml2name)
        Argument_Filter_NJ
    XMLor XMLdo (STR ''usableRulesNonJoin'') {
      a \<leftarrow>? xml_leaf (STR ''left'') True XMLor xml_leaf (STR ''right'') False;
      case a of None \<Rightarrow> XMLdo {
          b \<leftarrow> xml2non_join_info xml2name;
          xml_return (Usable_Rules_Reach_NJ b)
        }
      | Some left \<Rightarrow> XMLdo {
          U \<leftarrow> xml_singleton (STR ''usableRules'') (XTC_Parser.rules xml2name) id;
          p \<leftarrow> xml2non_join_info xml2name;
          xml_return (Usable_Rules_Reach_Unif_NJ (if left then Inl U else Inr U) p)
        }
    }
  ) x"

partial_function (sum_bot) xml2ncomm_proof :: "(string,label_type)lab xmlt2 \<Rightarrow> (string, label_type,string,string) ncomm_proof xmlt2"
where [code]: "xml2ncomm_proof xml2name x = (let rew = CPF_Parser.rsteps xml2name in
    xml_singleton (STR ''comDisproof'') (
      xml_triple (STR ''nonJoinableFork'') rew rew
           (xml2non_join_info xml2name)
           (\<lambda> (s,seq1) (_,seq2). Non_Join_Comm s seq1 seq2)
  ) id x)"

partial_function (sum_bot) xml2comm_proof :: "((string,label_type)lab,string) comm_proof xmlt2"
  where [code]: "xml2comm_proof x = xml_singleton (STR ''comProof'') (       
      xml_change (xml_do (STR ''parallelClosed'') nat_or_empty) (xml_return \<circ> Parallel_Closed_Comm)
    XMLor 
      xml_singleton (STR ''swapTRSs'') xml2comm_proof Swap_Comm
    ) id x"

partial_function (sum_bot) xml2ncr_proof :: "(string,label_type)lab xmlt2 \<Rightarrow> (string, label_type,string,string) ncr_proof xmlt2"
where [code]: "xml2ncr_proof xml2name x = (let rew = CPF_Parser.rsteps xml2name in
    xml_singleton (STR ''crDisproof'') (
      XMLdo (STR ''nonWcrAndSN'') {
       _ \<leftarrow> xml_any;
       prf \<leftarrow> xml2trs_termination_proof xml2name;
       xml_return (SN_NWCR prf)
    } XMLor xml_triple (STR ''nonJoinableFork'') rew rew
           (xml2non_join_info xml2name)
           (\<lambda> (s,seq1) (_,seq2). Non_Join s seq1 seq2)
    XMLor xml_pair (STR ''modularityDisjoint'')
        (xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id)
        (xml2ncr_proof xml2name)
        NCR_Disj_Subtrs
    XMLor xml_triple (STR ''redundantRules'')
        (xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id)
        (xml_nat (STR ''nrSteps''))
        (xml2ncr_proof xml2name)
      NCR_Redundant_Rules
    XMLor xml_pair (STR ''persistentDecomposition'')
        (xml2ms_signature xml2name)
        (xml_pair (STR ''component'')
            (xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id)
            (xml2ncr_proof xml2name)
            (\<lambda> trs prf. (trs, prf)))
       (\<lambda> sig (trs, prf). NCR_Persistent_Decomposition sig trs prf)
  ) id x)"

hide_const(open) unraveling_info

partial_function (sum_bot) xml2conditional_ncr_proof ::
    "(string, label_type) lab xmlt2 \<Rightarrow> (string, label_type, string, string) conditional_ncr_proof xmlt2"
where [code]: "xml2conditional_ncr_proof xml2name x = xml_singleton (STR ''conditionalCrDisproof'') (
    xml_singleton (STR ''unconditional'')
      (CPF_Parser.xml2ncr_proof xml2name)
      Unconditional_CNCR
    XMLor xml_triple (STR ''inlineConditions'')
      (XMLdo (STR ''rules'') {ret \<leftarrow>* XTC_Parser.crule xml2name (STR ''rule''); xml_return ret})
      (xml2inline_cond_info xml2name)
      (xml2conditional_ncr_proof xml2name)
      (Transformation_CNCR \<circ>\<circ> Inline_Conditions_CCRT)
    XMLor xml_pair (STR ''infeasibleRuleRemoval'')
      (xml2infeasible_rules_info xml2name)
      (xml2conditional_ncr_proof xml2name)
      (Transformation_CNCR \<circ> Infeasible_Rule_Removal_CCRT)
    XMLor xml_tuple4 (STR ''nonJoinableFork'')
      (xml_triple (STR ''terms'')
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
        (XTC_Parser.term xml2name)
      (\<lambda>s t u. (s, t, u)))
      (CPF_Parser.csteps xml2name)
      (CPF_Parser.csteps xml2name)
      (xml2non_join_info xml2name)
      (\<lambda>(s, t, u) ps qs i. Non_Join_CNCR s t u ps qs i)
  ) id x"

partial_function (sum_bot) xml2eq_proof :: "'a::showl xmlt2 \<Rightarrow> ('a, string) eq_proof xmlt2" where
  [code]: "xml2eq_proof xml2name x = (
    xml_singleton (STR ''refl'') (XTC_Parser.term xml2name) Equational_Reasoning.Refl
    XMLor xml_singleton (STR ''sym'') (xml2eq_proof xml2name) Sym
    XMLor xml_pair (STR ''trans'') (xml2eq_proof xml2name) (xml2eq_proof xml2name) Trans
    XMLor xml_pair (STR ''assm'') (XTC_Parser.rule xml2name) (CPF_Parser.subst xml2name) (\<lambda>r s. Assm r (mk_subst Var s))
    XMLor XMLdo (STR ''cong'') {
      a \<leftarrow> xml2name;
      b \<leftarrow>* xml2eq_proof xml2name;
      xml_return (Cong a b)
   }) x"

definition
  xml2equational_proof :: "('f :: {compare_order,showl}, label_type) lab xmlt2 \<Rightarrow> ('f, label_type, string) equational_proof xmlt2"
where
  "xml2equational_proof xml2name =
    xml_singleton (STR ''equationalProof'') (
      xml_singleton (STR ''equationalProofTree'') (xml2eq_proof xml2name) Equational_Proof_Tree
      XMLor xml_change (CPF_Parser.conversion xml2name) (\<lambda>(_, e). xml_return (Conversion e))
      XMLor xml_change (CPF_Parser.subsumption_proof xml2name) (xml_return \<circ> Conversion_With_History)
      XMLor xml_pair (STR ''completionAndNormalization'')
        (xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id)
        (xml2completion_proof xml2name)
        Completion_and_Normalization
    ) id"

definition xml2equational_disproof ::
    "('f :: {compare_order,showl}, label_type) lab xmlt2 \<Rightarrow> ('f, label_type, string) equational_disproof xmlt2"
  where
    "xml2equational_disproof xml2name =
      xml_singleton (STR ''equationalDisproof'') (
        (xml_pair (STR ''completionAndNormalization'')
          (xml_singleton (STR ''trs'') (XTC_Parser.rules xml2name) id)
          (xml2completion_proof xml2name)
          Completion_and_Normalization_Different)
        XMLor xml_change (xml2ordered_completion xml2name) (\<lambda>(rs, es, ro, p).
          xml_return (Ordered_Completion_and_Normalization_Different rs es ro p))
        XMLor (xml_singleton (STR ''convertibleInstance'') (CPF_Parser.subsumption_proof xml2name) Convertible_Instance)
      ) id"

type_synonym lsymbol = "(string, label_type) lab"

definition xml2cert_problem :: "string xmlt2 \<Rightarrow> lsymbol xmlt2 \<Rightarrow> _ xmlt2"
where
  "xml2cert_problem xml2uname xml2name = XMLdo (STR ''certificationProblem'') {
    inp \<leftarrow>? CPF_Parser.xml2input xml2name;
    version \<leftarrow>? xml_text (STR ''cpfVersion'');
    p \<leftarrow> XMLdo (STR ''proof'') {
      p \<leftarrow> xml_change (xml2trs_termination_proof xml2name) (xml_return \<circ> TRS_Termination_Proof)
        XMLor xml_change (xml2trs_nontermination_proof xml2name) (xml_return \<circ> TRS_Nontermination_Proof)
        XMLor xml_change (xml2reltrs_nontermination_proof xml2name) (xml_return \<circ> Relative_TRS_Nontermination_Proof)
        XMLor xml_change (xml2cr_proof xml2name) (xml_return \<circ> TRS_Confluence_Proof)
        XMLor xml_change (xml2ncr_proof xml2name) (xml_return \<circ> TRS_Non_Confluence_Proof)
        XMLor xml_change (xml2dp_termination_proof xml2name) (xml_return \<circ> DP_Termination_Proof)
        XMLor xml_change (xml2dp_nontermination_proof xml2name) (xml_return \<circ> DP_Nontermination_Proof)
        XMLor xml_change (xml2completion_proof xml2name) (xml_return \<circ> Completion_Proof)
        XMLor xml_change (xml2ordered_completion_proof xml2name) (xml_return \<circ> Ordered_Completion_Proof)
        XMLor xml_change (xml2equational_proof xml2name) (xml_return \<circ> Equational_Proof)
        XMLor xml_change (xml2equational_disproof xml2name) (xml_return \<circ> Equational_Disproof)
        XMLor xml_change (xml2complexity_proof xml2name) (xml_return \<circ> Complexity_Proof)
        XMLor xml_change (xml2quasi_reductive_proof xml2name) (xml_return \<circ> Quasi_Reductive_Proof)
        XMLor xml_change (xml2conditional_cr_proof xml2name) (xml_return \<circ> Conditional_CR_Proof)
        XMLor xml_change (xml2conditional_ncr_proof xml2name) (xml_return \<circ> Conditional_Non_CR_Proof)
        XMLor xml_change (xml2ncomm_proof xml2name) (xml_return \<circ> TRS_Non_Commutation_Proof)
        XMLor xml_change xml2comm_proof (xml_return \<circ> TRS_Commutation_Proof)
        XMLor xml_singleton (STR ''treeAutomatonClosedProof'') closed_criterion Tree_Automata_Closed_Proof
        XMLor xml_change (xml2ac_termination_proof xml2name) (xml_return \<circ> AC_Termination_Proof)
        XMLor xml_singleton (STR ''ltsTerminationProof'') lts_termination_proof_parser LTS_Termination_Proof
        XMLor xml_singleton (STR ''ltsSafetyProof'') lts_safety_proof_parser LTS_Safety_Proof
        XMLor xml_change (xml2infeasibility_proof xml2name) (xml_return \<circ> Infeasibility_Proof)
        XMLor xml_change (xml2unknown_proof xml2name) (xml_return \<circ> Unknown_Proof)
        XMLor xml_change (xml2unknown_disproof xml2name) (xml_return \<circ> Unknown_Disproof);
      xml_return p
    };
    orig \<leftarrow>? xml_any;
    xml_return (inp, p)
  }"

partial_function (sum_bot) plain_name :: "lsymbol xmlt2"
where [code]: "plain_name x = (
    xml_change (xml_text (STR ''name'')) (xml_return \<circ> UnLab)
    XMLor xml_singleton (STR ''sharp'') plain_name Sharp
    XMLor xml_pair (STR ''labeledSymbol'') plain_name (
        xml_many (STR ''numberLabel'') (xml_nat (STR ''number'')) Inl
        XMLor xml_many (STR ''symbolLabel'') plain_name Inr
      )
      (\<lambda>f l. case l of
        Inl ls  \<Rightarrow> Lab f ls
      | Inr ls \<Rightarrow> FunLab f ls)
  ) x"

definition "parse_cert_problem =
  parse_xmlfile (
    debug (showsl (STR ''0'')) (STR ''parsing xml to certification problem'')
    (xml2cert_problem (xml_text (STR ''name'')) CPF_Parser.plain_name))"

definition xml_symbols :: "string \<Rightarrow> lsymbol list \<Rightarrow> xml" where
  "xml_symbols tagname Fs = XML tagname [] []"

definition xml_equation_literal :: "(_,_) equation_literal \<Rightarrow> xml" where
  "xml_equation_literal eql =
    (case eql of
      Equation (s,t) \<Rightarrow> XML ''equation'' [] [xml_term s, xml_term t]
    | Inequality (s,t) \<Rightarrow> XML ''inequality'' [] [xml_term s, xml_term t])"

primrec xml_input :: "_ \<Rightarrow> xml" where
  "xml_input (Inn_TRS_input Q R S start) = xml_trs_input (Inl Q) R S"
| "xml_input (FP_TRS_input Q R) = xml_trs_input (Inr Q) R []"
| "xml_input (DP_input m P Q R) =
    XML ''dpInput'' [] (
      xml_rules ''trs'' R #
      xml_rules ''dps'' P #
      xml_strategy (Inl Q) @ [XML ''minimal'' [] [XML_text (if m then ''true'' else ''false'')]])"
| "xml_input (COMP_input E R) =
    XML ''completionInput'' [] [xml_rules ''equations'' E, xml_rules ''trs'' R]"
| "xml_input (OCOMP_input E\<^sub>0 E R ro) =
    XML ''orderedCompletionInput'' [] [xml_rules ''equations'' E\<^sub>0, XML ''orderedCompletionResult'' [] [
      xml_rules ''trs'' R, xml_rules ''equations'' E, XML ''reductionOrder'' [] [xml_redord ro]]]"
| "xml_input (EQ_input E eql) =
    XML ''equationalReasoningInput'' [] [xml_rules ''equations'' E, xml_equation_literal eql]"
| "xml_input (CPX_input Q R S cm cc) =
    XML ''complexityInput'' [] [xml_trs_input (Inl Q) R S, xml_complexity_measure cm, xml_complexity_class cc]"
| "xml_input (CTRS_input ctrs) = xml_ctrs_input ctrs"
| "xml_input (TA_input ta r) = xml_ta_input ta r"
| "xml_input (AC_input r a c) = xml_ac_input r a c"
| "xml_input (LTS_input r) = xml_lts r"
| "xml_input (LTS_safety_input r e) = xml_lts_safety r e"
| "xml_input (Commutation_input f r s) =
    XML ''commutationInput'' [] [xml_signature f, xml_rules ''trs'' r, xml_rules ''trs'' s]"
| "xml_input (Infeasibility_input c r) =
    XML ''infeasibilityInput'' [] [XML ''rules'' [] (map xml_crule c), xml_rules ''trs'' r]"
| "xml_input (Unknown_input u) = XML ''unknownInput'' [] [XML_text u]"

hide_const (open) rstep estep cstep

end
